import arcgis
from arcgis import gis
from .admin.administration import Server
import logging
_log = logging.getLogger(__name__)
###########################################################################
class ServerManager(object):
    """
    Helper class for managing your ArcGIS Servers. This class is not created
    by users directly. An instance of this class, called 'servers',
    is available as a property of the gis.admin object. Administrators call methods
    on this 'gis.admin.servers' object to manage and interrogate ArcGIS Servers.
    """
    _gis = None
    _catalog_list = None
    _server_list = None
    _portal = None
    _gis = None
    _pa = None
    _federation = None

    #----------------------------------------------------------------------
    def __init__(self, gis):
        self._gis = gis
        self._portal = gis._portal
        self._pa = gis.admin
        self._federation = self._pa.federation
        self._server_list = None
        self._catalog_list = None
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._pa._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._pa._url)
    #----------------------------------------------------------------------
    def list(self):
        """
        Retrieves all servers in a GIS.

        :returns:
           A list of all servers found in the GIS.
        """
        from . import ServicesDirectory
        if self._server_list is not None:
            return self._server_list

        self._server_list = []
        self._catalog_list = []
        res = self._portal.con.post("portals/self/servers", {"f": "json"})
        servers = res['servers']
        admin_url = None
        for server in servers:
            try:
                admin_url = server['adminUrl']
                if server['serverFunction'] == 'NotebookServer':
                     self._server_list.append(
                          Server(url=admin_url, gis=self._gis)
                     )
                else:
                      c = ServicesDirectory(url=admin_url, portal_connection=self._gis._portal.con, )
                      self._server_list.append(c.admin)
                      self._catalog_list.append(c)
            except:
                _log.warning("Could not access the server at " + admin_url)

        return self._server_list
    #----------------------------------------------------------------------
    def get(self, role=None, function=None):
        """
        Retrieves the ArcGIS Server(s) by role or function. While each argument is optional, 
        at least one argument must be set with an allowed value other than None.
        
        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        role                   Optional string. Limits the returned ArcGIS Servers based on the 
                               server's role as either a hosting server for the portal, a federated server, 
                               or a server with restricted access to publishing. The allowed values 
                               are HOSTING_SERVER, FEDERATED_SERVER, or FEDERATED_SERVER_WITH_RESTRICTED_PUBLISHING, 
                               respectively.
        ------------------     --------------------------------------------------------------------
        function               Optional string. Limits the returned ArcGIS Servers based on the  
                               server's function. Provide a comma-separated list of values. The 
                               allowed values are GeoAnalytics, RasterAnalytics, and ImageHosting.
        ==================     ====================================================================

        :return:
           The ArcGIS Server(s) discovered that match the criteria. 
        """
        servers = []
        if role is None and function is None:
            raise ValueError("A role or function must be provided")
        for server in self._federation.servers['servers']:
            if str(role).lower() == server['serverRole'].lower():
                servers.append(Server(url=server['adminUrl'], gis=self._gis))
            elif str(function).lower() in server['serverFunction'].lower():
                servers.append(Server(url=server['adminUrl'], gis=self._gis))
        return servers
    #----------------------------------------------------------------------
    @property
    def _server_info(self):
        """
        Gets federation information for all servers associated the WebGIS.
        """
        return self._federation.servers['servers']
    #----------------------------------------------------------------------
    def _federate(self,
                 url,
                 admin_url,
                 username,
                 password):
        """
        This operation enables ArcGIS Servers to be federated with Portal
        for ArcGIS.
        
        .. note::
            For the url argument - If the site includes the Web Adaptor, 
            the URL includes the Web Adaptor address, for example, 
            https://webadaptor.domain.com/arcgis. If you've added ArcGIS Server 
            to your organization's reverse proxy server, the URL is the reverse 
            proxy server address (for example, 
            https://reverseproxy.domain.com/myorg). Note that the federation 
            operation will perform a validation check to determine if the 
            provided URL is accessible from the server site. If the resulting 
            validation check fails, a warning will be generated in the Portal 
            for ArcGIS logs. However, federation will not fail if the URL is 
            not validated, as the URL may not be accessible from the server 
            site, such as is the case when the server site is behind a firewall.
          
        
        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The URL of the GIS server used by external users when 
                               accessing the ArcGIS Server site.  See note above.
        ------------------     --------------------------------------------------------------------
        admin_url              Required string. The URL used for accessing ArcGIS Server when 
                               performing administrative operations on the internal network, for 
                               example, https://gisserver.domain.com:6443/arcgis.
        ------------------     --------------------------------------------------------------------
        username               Required string. The username of the primary site administrator account.
        ------------------     --------------------------------------------------------------------
        password               Required integer. The password of the username above.
        ==================     ====================================================================
        

        :return:
           The server response with server ID.
        """
        res = self._federation.federate(url,
                                        admin_url,
                                        username,
                                        password)
        self._server_list = None
        return res
    #----------------------------------------------------------------------
    def _unfederate(self, server_id):
        """
        This operation unfederates an ArcGIS Server from Portal for ArcGIS.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        server_id              Required string. The unique ID of the server.
        ==================     ====================================================================
        
        
        :return:
           The server response with server ID.
        """
        res = self._federation(server_id)
        self._server_list = None
        return res
    #----------------------------------------------------------------------
    def update(self, server, role, function=None):
        """
        This operation allows you to set an ArcGIS Server federated with
        Portal for ArcGIS as the hosting server or to enforce fine-grained
        access control to a federated server. You can also remove hosting
        server status from an ArcGIS Server. To set a hosting server, an
        enterprise geodatabase must be registered as a managed database
        with the ArcGIS Server.
        
        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        server                 Required string. The arcgis.gis.Server object.
        ------------------     --------------------------------------------------------------------
        role                   Required string. State whether the server is either a hosting server 
                               for the portal, a federated server, or a server with restricted access 
                               to publishing. The allowed values are HOSTING_SERVER, FEDERATED_SERVER, 
                               or FEDERATED_SERVER_WITH_RESTRICTED_PUBLISHING, respectively.
        ------------------     --------------------------------------------------------------------
        function               Optional string. The specific function associated with this server. Provide a 
                               comma-separated list of values, but it is not recommend that a single
                               server have all the server functions. The allowed values are GeoAnalytics, 
                               RasterAnalytics, and ImageHosting. 
        ==================     ====================================================================


        :return:
           A status message of 'success' with the ID of the ArcGIS Server.
        """
        if isinstance(server, Server) == False:
            raise ValueError("server must be of type arcgis.gis.Server")
        roles = ["FEDERATED_SERVER",
                 "FEDERATED_SERVER_WITH_RESTRICTED_PUBLISHING",
                 "HOSTING_SERVER"]
        functions = {
            "geoanalytics" : "GeoAnalytics",
            "rasteranalytics" : "RasterAnalytics",
            "imagehosting" : "ImageHosting",
            "none" : None
        }
        if not role.upper() in roles:
            raise ValueError("Invalid role, allowed values: %s" % ",".join(roles))
        if not str(function).lower() in functions.keys():
            raise ValueError("Invalid function, allowed values: %s" % ",".join(functions.keys()))
        else:
            function = functions[str(function).lower()]
        server_id = None
        from six.moves.urllib.parse import urlparse
        b = urlparse(url=server._admin_url).netloc.lower()
        for s in self._server_info:
            if b == urlparse(s['adminUrl'].lower()).netloc:
                server_id = s['id']
                break
            del s
        return self._federation.update(server_id, role, function)
    #----------------------------------------------------------------------
    def validate(self):
        """
        This operation returns information on the status of ArcGIS Servers
        registered with Portal for ArcGIS.

        :return:
           True if all servers are functioning as expected, False if there is an 
           issue with 1 or more of the Federated Servers.  
        """
        return self._federation.validate_all()['status'] == 'success'


"""
This resource is a collection of all the usage reports created within
your site. The Create Usage Report operation lets you define a new
usage report.
"""
from __future__ import absolute_import
from __future__ import print_function
import json
import six
from .._common import BaseServer

########################################################################
class ReportManager(BaseServer):
    """
    A utility class for managing usage reports for ArcGIS Server.

    """
    _con = None
    _json_dict = None
    _url = None
    _json = None
    _metrics = None
    _reports = None
    #----------------------------------------------------------------------
    def __init__(self,
                 url,
                 gis,
                 initialize=False):
        """Constructor

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        gis                    Optional string. The GIS or Server object.
        ------------------     --------------------------------------------------------------------
        initialize             Optional string. Denotes whether to load the machine properties at
                               creation (True). Default is False.
        ==================     ====================================================================

        """
        super(ReportManager, self).__init__(url=url, gis=gis)
        if url.lower().endswith('/usagereports'):
            self._url = url
        else:
            self._url = url + "/usagereports"
        self._con = gis
        if initialize:
            self._init(gis)
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def list(self):
        """Retrieves a list of reports on the server.

        :return:
            A list of reports found.

        """
        if self.properties is None:
            self._init()
        self._reports = []
        if isinstance(self.properties['metrics'], list):
            for r in self.properties['metrics']:
                url = self._url + "/%s" % six.moves.urllib.parse.quote(r['reportname'])
                self._reports.append(Report(url=url,
                                            gis=self._con))
                del url
        return self._reports
    #----------------------------------------------------------------------
    @property
    def settings(self):
        """
        Gets the current usage reports settings. The usage reports
        settings are applied to the entire site. When usage
        reports are enabled, service usage statistics are collected and
        persisted to a statistics database. When usage reports are
        disabled, the statistics are not collected. The interval
        parameter defines the duration (in minutes) during which the usage
        statistics are sampled or aggregated (in-memory) before being
        written out to the statistics database. Database entries are
        deleted after the interval specified in the max_history parameter (
        in days), unless the max_history parameter is 0, for which the
        statistics are persisted forever.
        """
        params = {
            "f" : "json"
        }
        url = self._url + "/settings"
        return self._con.get(path=url,
                             params=params)
    #----------------------------------------------------------------------
    def edit(self,
             interval,
             enabled=True,
             max_history=0):
        """
        Edits the usage reports settings that are applied to the entire site.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        interval               Required string. Defines the duration (in minutes) for which the
                               usage statistics are aggregated or sampled, in-memory, before being
                               written out to the statistics database.
        ------------------     --------------------------------------------------------------------
        enabled                Optional string. When usage reports are enabled, service
                               usage statistics are collected and persisted to a statistics
                               database. When usage reports are disabled, the statistics are not
                               collected.  The default is True (enabled).
        ------------------     --------------------------------------------------------------------
        max_history            Optional integer. The number of days after which usage statistics
                               are deleted from the statistics database. If the max_history
                               parameter is set to 0 (the default value), the statistics are
                               persisted forever.
        ==================     ====================================================================


        :return:
            A JSON message indicating success.

        """
        params = {
            "f" : "json",
            "maxHistory" : max_history,
            "enabled" : enabled,
            "samplingInterval" : interval
        }
        url = self._url + "/settings/edit"
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def create(self,
               reportname,
               queries,
               metadata=None,
               since="LAST_DAY",
               from_value=None,
               to_value=None,
               aggregation_interval=None):
        """
        Creates a new usage report. A usage report is created by submitting
        a JSON representation of the usage report to this operation.

        ====================     ====================================================================
        **Argument**             **Description**
        --------------------     --------------------------------------------------------------------
        reportname               Required string. The unique name of the report.
        --------------------     --------------------------------------------------------------------
        queries                  Required string. A list of queries for which to generate the report.
                                 Specify the list as an array of JSON objects representing the queries.
                                 Each query specifies the list of metrics to be queried for a given
                                 set of resourceURIs.

                                 The queries parameter has the following sub-parameters:

                                 - resourceURIs -- Comma-separated list of resource URIs for which
                                   to report metrics. This specifies the services or folders for
                                   which to gather metrics. The resourceURI is formatted as below:
                                    - services/ -- Entire Site
                                    - services/Folder/ -- Folder within a Site. Reports metrics
                                      aggregated across all services within that Folder and Sub-Folders.
                                    - services/Folder/ServiceName.ServiceType -- Service in a
                                      specified folder, for example:
                                         - services/Map_bv_999.MapServer
                                         - services/ServiceName.ServiceType
                                    - Service in the root folder, for example: Map_bv_999.MapServer.
        --------------------     --------------------------------------------------------------------
        metadata                 Optional string. Any JSON object representing presentation tier
                                 data for the usage report, such as report title, colors,
                                 line-styles, etc. Also used to denote visibility in ArcGIS Server
                                 Manager for reports created with the Administrator Directory. To
                                 make any report created in the Administrator Directory visible to
                                 Manager, include *"managerReport":true* in the metadata JSON object.
                                 When this value is not set (default), reports are not visible in
                                 Manager. This behavior can be extended to any client that wants to
                                 interact with the Administrator Directory. Any user-created value
                                 will need to be processed by the client.
        --------------------     --------------------------------------------------------------------
        since                    Optional string. The time duration of the report. The supported
                                 values are: LAST_DAY, LAST_WEEK, LAST_MONTH, LAST_YEAR, CUSTOM

                                 - LAST_DAY represents a time range spanning the previous 24 hours.
                                   This is the default value.
                                 - LAST_WEEK represents a time range spanning the previous 7 days.
                                 - LAST_MONTH represents a time range spanning the previous 30 days.
                                 - LAST_YEAR represents a time range spanning the previous 365 days.
                                 - CUSTOM represents a time range that is specified using the from
                                   and to parameters.
        --------------------     --------------------------------------------------------------------
        from_value               Optional string. Only valid when *since* is CUSTOM. The timestamp
                                 (milliseconds since UNIX epoch, namely January 1, 1970, 00:00:00 GMT)
                                 for the beginning period of the report.
        --------------------     --------------------------------------------------------------------
        to_value                 Optional string. Only valid when *since* is CUSTOM. The timestamp
                                 (milliseconds since UNIX epoch, namely January 1, 1970, 00:00:00 GMT)
                                 for the ending period of the report.
        --------------------     --------------------------------------------------------------------
        aggregation_interval     Optional string. The aggregation interval in minutes. Server metrics
                                 are aggregated and returned for time slices aggregated using the
                                 specified aggregation interval. The time range for the report,
                                 specified using the *since* parameter (and *from_value* and
                                 *to_value* when since is CUSTOM) is split into multiple slices, each
                                 covering an aggregation interval. Server metrics are then aggregated
                                 for each time slice and returned as data points in the report data.
                                 When the aggregation_interval is not specified, the following defaults
                                 are used:

                                   - LAST_DAY: 30 minutes
                                   - LAST_WEEK: 4 hours
                                   - LAST_MONTH: 24 hours
                                   - LAST_YEAR: 1 week
                                   - CUSTOM: 30 minutes up to 1 day, 4 hours up to 1 week, 1
                                   day up to 30 days, and 1 week for longer periods.

                                 If the interval specified in Usage Reports Settings is more than
                                 the aggregationInterval, the interval is used instead.
        ====================     ====================================================================


        :return:
            A JSON indicating success.


        .. code-block:: python

            USAGE EXAMPLE:

            >>> queryObj = [{
                "resourceURIs": ["services/Map_bv_999.MapServer"],
                "metrics": ["RequestCount"]
                }]
            >>> obj.createReport(
                reportname="SampleReport",
                queries=queryObj,
                metadata="This could be any String or JSON Object.",
                since="LAST_DAY"
                )
        """
        url = self._url + "/add"
        temp = False
        params = {

            "reportname" : reportname,
            "since" : since,
        }
        if  not metadata:
            params['metadata'] = {
                "temp" : temp,
                "title" : reportname,
                "managerReport" : False,

            }
        else:
            params['metadata'] = metadata
        if isinstance(queries, dict):
            params["queries"] = [queries]
        elif isinstance(queries, list):
            params["queries"] = queries
        if aggregation_interval:
            params['aggregationInterval'] = aggregation_interval
        if since.lower() == "custom":
            params['to'] = to_value
            params['from'] = from_value
        p = {"f" : "json",'usagereport' : params}
        res = self._con.post(path=url,
                             postdata=p)
        #  Refresh the metrics object
        self._init()
        for report in self.list():
            if report.reportname.lower() == reportname.lower():
                return report
        return res
    #----------------------------------------------------------------------
    def quick_report(self,
                     since="LAST_WEEK",
                     queries="services/",
                     metrics="RequestsFailed"):
        """
        The operation quick_report generates an on the fly usage report for
        a service, services, or folder.

        ====================     ====================================================================
        **Argument**             **Description**
        --------------------     --------------------------------------------------------------------
        since                    Optional string. The time duration of the report. The supported
                                 values are: LAST_DAY, LAST_WEEK, LAST_MONTH, LAST_YEAR, CUSTOM

                                 - LAST_DAY represents a time range spanning the previous 24 hours.
                                   This is the default value.
                                 - LAST_WEEK represents a time range spanning the previous 7 days.
                                 - LAST_MONTH represents a time range spanning the previous 30 days.
                                 - LAST_YEAR represents a time range spanning the previous 365 days.
                                 - CUSTOM represents a time range that is specified using the from
                                   and to parameters.
        --------------------     --------------------------------------------------------------------
        queries                  Required string. A list of queries for which to generate the report.
                                 Specify the list as an array of JSON objects representing the queries.
                                 Each query specifies the list of metrics to be queried for a given
                                 set of resourceURIs.

                                 The queries parameter has the following sub-parameters:

                                 - resourceURIs -- Comma-separated list of resource URIs for which
                                   to report metrics. This specifies the services or folders for
                                   which to gather metrics. The resourceURI is formatted as below:
                                    - services/ -- Entire Site
                                    - services/Folder/ -- Folder within a Site. Reports metrics
                                      aggregated across all services within that Folder and Sub-Folders.
                                    - services/Folder/ServiceName.ServiceType -- Service in a
                                      specified folder, for example:
                                         - services/Map_bv_999.MapServer
                                         - services/ServiceName.ServiceType
                                    - Service in the root folder, for example: Map_bv_999.MapServer.
        --------------------     --------------------------------------------------------------------
        metrics                  Optional string. Comma separated list of metrics to be reported.

                                 Supported metrics are:

                                    - RequestCount -- the number of requests received
                                    - RequestsFailed -- the number of requests that failed
                                    - RequestsTimedOut -- the number of requests that timed out
                                    - RequestMaxResponseTime -- the maximum response time
                                    - RequestAvgResponseTime -- the average response time
                                    - ServiceActiveInstances -- the maximum number of active
                                      (running) service instances sampled at 1 minute intervals,
                                      for a specified service
        ====================     ====================================================================


        :return:
            A Python dictionary of data on a successful query.

        """
        from uuid import uuid4
        queries = {
            "resourceURIs": queries.split(','),
            "metrics" : metrics.split(',')
        }
        reportname = uuid4().hex
        metadata = {
                "temp" : True,
                "title" : reportname,
                "managerReport" : False,

            }
        res = self.create(reportname=reportname,
                                       queries=queries,
                                       since=since,
                                       metadata=metadata)
        if isinstance(res, Report):
            data = res.query()
            res.delete()
            return data
        return res
########################################################################
class Report(BaseServer):
    """
    **(This class should not be created by a user)**

    A utility class representing a single usage report returned by ArcGIS Server.

    A Usage Report is used to obtain ArcGIS Server usage data for specified
    resources during a given time period. It specifies the parameters for
    obtaining server usage data, time range (parameters since, from_value, to_value),
    aggregation interval, and queries (which specify the metrics to be
    gathered for a collection of server resources, such as folders and
    services).
    """
    _con = None
    _url = None
    _json = None
    _reportname = None
    _since = None
    _from = None
    _to = None
    _aggregationInterval = None
    _queries = None
    _metadata = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis,
                 initialize=False):
        """
        Constructor

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        gis                    Optional string. The GIS or Server object.
        ------------------     --------------------------------------------------------------------
        initialize             Optional string. Denotes whether to load the machine properties at
                               creation (True). Default is False.
        ==================     ====================================================================

        """
        super(Report, self).__init__(url=url, gis=gis)
        self._con = gis
        self._url = url
        if initialize:
            self._init()
    #----------------------------------------------------------------------
    def edit(self):
        """
        Edits the usage report. To edit a usage report, submit
        the complete JSON representation of the usage report which
        includes updates to the usage report properties. The name of the
        report cannot be changed when editing the usage report.

        Values are changed in the class, to edit a property like
        metrics, pass in a new value.

        :return:
            A JSON indicating success.

        """

        usagereport_dict = {
            "reportname": self._reportname,
            "queries": self._queries,
            "since": self._since,
            "metadata": self._metadata,
            "to" : self._to,
            "from" : self._from,
            "aggregationInterval" : self._aggregationInterval
        }
        params = {
            "f" : "json",
            "usagereport" : json.dumps(usagereport_dict)
        }
        url = self._url + "/edit"
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def delete(self):
        """
        Deletes this usage report.

        :return:
            A JSON indicating success.
        """
        url = self._url + "/delete"
        params = {
            "f" : "json",
        }
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def query(self, query_filter=None):
        """
        Retrieves server usage data for this report. This operation
        aggregates and filters server usage statistics for the entire
        ArcGIS Server site. The report data is aggregated in a time slice,
        which is obtained by dividing up the time duration by the default
        (or specified) aggregationInterval parameter in the report. Each
        time slice is represented by a timestamp, which represents the
        ending period of that time slice.

        In the JSON response, the queried data is returned for each metric-
        resource URI combination in a query. In the report-data section,
        the queried data is represented as an array of numerical values. A
        response of null indicates that data is not available or requests
        were not logged for that metric in the corresponding time-slice.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        query_filter           Optional string. The report data can be filtered by the machine
                               where the data is generated. The filter accepts a comma-separated
                               list of machine names; * represents all machines.
        ==================     ====================================================================


        :return:
            A JSON containing the server usage data.


        .. code-block:: python

            USAGE EXAMPLE 1: Filters for the specified machines

            {"machines": ["WIN-85VQ4T2LR5N", "WIN-239486728937"]}

        .. code-block:: python

            USAGE EXAMPLE 2: No filtering, all machines are accepted

            {"machines": "*"}

        """
        if query_filter is None:
            query_filter = {"machines": "*"}
        params = {
            "f" : "json",
            "filter" : query_filter,
            "filterType" : 'json'
        }
        url = self._url + "/data"
        return self._con.get(path=url, params=params)



""" The portalpy module for working with the ArcGIS Online and Portal APIs."""
from __future__ import absolute_import
import io
import os
import re
import ssl
import sys
import json
import uuid
import zlib
import shutil
import logging
import tempfile
import mimetypes
import unicodedata
try:
    #PY2
    from cStringIO import StringIO
except ImportError:
    #PY3
    from io import StringIO
from io import BytesIO
from collections import OrderedDict

import six
from six.moves.urllib_parse import urlparse, urlunparse, parse_qsl
from six.moves.urllib_parse import quote, unquote, urlunsplit
from six.moves.urllib_parse import urlencode, urlsplit
from six.moves.urllib.error import HTTPError
from six.moves.urllib import request
from six.moves import http_cookiejar as cookiejar
from six.moves import http_client
from .common._utils import Error
__version__ = '1.6.0'
_log = logging.getLogger(__name__)

DEFAULT_TOKEN = uuid.uuid4()

########################################################################
class MultiPartForm(object):
    """Accumulate the data to be used when posting a form."""
    PY2 = sys.version_info[0] == 2
    PY3 = sys.version_info[0] == 3
    files = []
    form_fields = []
    boundary = None
    form_data = ""
    #----------------------------------------------------------------------
    def __init__(self, param_dict=None, files=None):
        if param_dict is None:
            param_dict = {}
        if files is None:
            files = {}
        self.boundary = None
        self.files = []
        self.form_data = ""
        if len(self.form_fields) > 0:
            self.form_fields = []

        if len(param_dict) == 0:
            self.form_fields = []
        else:
            for k,v in param_dict.items():
                self.form_fields.append((k,v))
                del k,v
        if isinstance(files, list):
            if len(files) == 0:
                self.files = []
            else:
                for key, filePath, fileName in files:
                    self.add_file(fieldname=key,
                                  filename=fileName,
                                  filePath=filePath,
                                  mimetype=None)
        elif isinstance(files, dict):
            for key, filepath in files.items():
                self.add_file(fieldname=key,
                              filename=os.path.basename(filepath),
                              filePath=filepath,
                              mimetype=None)
                del key, filepath
        self.boundary = "%s" % self._make_boundary()
    #----------------------------------------------------------------------
    def get_content_type(self):
        return 'multipart/form-data; boundary=%s' % self.boundary
    #----------------------------------------------------------------------
    def add_field(self, name, value):
        """Add a simple field to the form data."""
        self.form_fields.append((name, value))
    #----------------------------------------------------------------------
    def _make_boundary(self):
        """ creates a boundary for multipart post (form post)"""
        if six.PY2:
            return '----------------%s--' % uuid.uuid4().hex
        elif six.PY3:
            return '----------------%s--' % uuid.uuid4().hex
        else:
            from random import choice
            digits = "0123456789"
            letters = "abcdefghijklmnopqrstuvwxyz"
            return '----------------%s--'.join(choice(letters + digits) \
                                   for i in range(15))
    #----------------------------------------------------------------------
    def add_file(self, fieldname, filename, filePath, mimetype=None):
        """Add a file to be uploaded.
        Inputs:
           fieldname - name of the POST value
           fieldname - name of the file to pass to the server
           filePath - path to the local file on disk
           mimetype - MIME stands for Multipurpose Internet Mail Extensions.
             It's a way of identifying files on the Internet according to
             their nature and format. Default is None.
        """
        body = filePath
        if mimetype is None:
            mimetype = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
        self.files.append((fieldname, filename, mimetype, body))
    #----------------------------------------------------------------------
    @property
    def make_result(self):
        if self.PY2:
            self._2()
        elif self.PY3:
            self._3()
        return self.form_data
    #----------------------------------------------------------------------
    def _2(self):
        """python 2.x version of formatting body data"""
        boundary = self.boundary
        buf = StringIO()
        for (key, value) in self.form_fields:
            buf.write('--%s\r\n' % boundary)
            buf.write('Content-Disposition: form-data; name="%s"' % key)
            buf.write('\r\n\r\n%s\r\n' % value)
        for (key, filename, mimetype, filepath) in self.files:
            if os.path.isfile(filepath):
                buf.write('--{boundary}\r\n'
                          'Content-Disposition: form-data; name="{key}"; '
                          'filename="{filename}"\r\n'
                          'Content-Type: {content_type}\r\n\r\n'.format(
                              boundary=boundary,
                              key=key,
                              filename=filename,
                              content_type=mimetype))
                with open(filepath, "rb") as f:
                    shutil.copyfileobj(f, buf)
                buf.write('\r\n')
        buf.write('--' + boundary + '--\r\n\r\n')
        buf = buf.getvalue()
        self.form_data = buf
    #----------------------------------------------------------------------
    def _3(self):
        """ python 3 method"""
        boundary = self.boundary
        buf = BytesIO()
        textwriter = io.TextIOWrapper(
            buf, 'utf8', newline='', write_through=True)


        for(key, filename, mimetype, filepath) in self.files:
            if os.path.isfile(filepath):
                textwriter.write(
                    '--{boundary}\r\n'
                    'Content-Disposition: form-data; name="{key}"; '
                    'filename="{filename}"\r\n'
                    'Content-Type: {content_type}\r\n\r\n'.format(
                        boundary=boundary, key=key, filename=filename,
                        content_type=mimetype))
                with open(filepath, "rb") as f:
                    shutil.copyfileobj(f, buf)
                textwriter.write('\r\n')

        for (key, value) in self.form_fields:
            textwriter.write(
                '--{boundary}\r\n'
                'Content-Disposition: form-data; name="{key}"\r\n\r\n'
                '{value}\r\n'.format(
                    boundary=boundary, key=key, value=value))
        textwriter.write('--{}--\r\n\r\n'.format(boundary))
        self.form_data = buf.getvalue()
########################################################################
class HTTPSClientAuthHandler(request.HTTPSHandler):
    def __init__(self, key, cert):
        request.HTTPSHandler.__init__(self)
        self.key = key
        self.cert = cert
    def https_open(self, req):
        #Rather than pass in a reference to a connection class, we pass in
        # a reference to a function which, for all intents and purposes,
        # will behave as a constructor
        return self.do_open(self.getConnection, req)
    def getConnection(self, host, timeout=300):
        return  http_client.HTTPSConnection(host,
                                            key_file=self.key,
                                            cert_file=self.cert,
                                            timeout=timeout)
########################################################################
def jsonize_dict(val):
    if isinstance(val, (dict, list)):
        return json.dumps(val)
    else:
        return val

class _ArcGISConnection(object):
    """ A class users to manage connection to ArcGIS services (Portal and Server). """
    baseurl = None
    key_file = None
    cert_file = None
    all_ssl = None
    proxy_host = None
    proxy_port = None
    _token = None
    _product = None
    _referer = None
    _useragent = None
    _parsed_org_url = None
    _username = None
    _password = None
    _auth = None
    _tokenurl = None
    _token = None
    _refresh_token = None   # oauth
    _server_token = None
    _connection = None
    _portal_connection = None
    _service_url = None
    _handlers = None

    #----------------------------------------------------------------------
    def __init__(self, baseurl=None, tokenurl=None, username=None,
                 password=None, key_file=None, cert_file=None,
                 expiration=60, all_ssl=False, referer=None,
                 proxy_host=None, proxy_port=None,
                 connection=None, verify_cert=True,
                 client_id=None):
        """ The _ArcGISConnection constructor. Requires URL and optionally username/password. """
        if baseurl is None:
            self._is_arcpy = False
        else:
            self._is_arcpy = baseurl.lower() == "pro"
            self._auth = "PRO"
        if self._is_arcpy:
            try:
                import arcpy
                baseurl = arcpy.GetActivePortalURL()
                self.baseurl = self._validate_url(url=baseurl)
            except ImportError:
                raise Error("Could not import arcpy")
        else:
            self.baseurl = baseurl
        self._tokenurl = tokenurl
        '''_normalize_url(baseurl)'''
        self._product = self._check_product()
        self.key_file = key_file
        self.cert_file = cert_file
        self.all_ssl = all_ssl
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.token = None
        self._server_token = None
        self._connection = connection # second connection
        self._useragent = 'geosaurus/' + __version__
        self._verify_cert = verify_cert

        # Setup the referer and user agent
        if baseurl:
            if not referer and \
               cert_file is None and \
               key_file is None:
                referer = "http"#urlparse(baseurl).netloc
            self._referer = referer
            parsed_url = urlparse(self.baseurl)
            self._parsed_org_url = urlunparse((parsed_url[0], parsed_url[1], "", "", "", ""))

        self._username = username
        self._password = password

        self._client_id = client_id

        if client_id is not None:
            self._auth = 'OAUTH'
        elif cert_file is not None and key_file is not None:
            self._auth = "PKI"
        elif username is not None and password is not None:
            self._auth = "BUILTIN" # or BASIC (LDAP) or DIGEST
        else:
            self._auth = "ANON" # or IWA (NTLM or Kerberos) (self.login sets this up)

        if cert_file is None and key_file is None:
            self.login(username, password, expiration, client_id)

    #----------------------------------------------------------------------
    def _validate_url(self, url):
        """ensures the base url has the /sharing/rest"""
        if self._is_arcpy:
            if not url[-1] == '/':
                url += '/'
            if url.lower().find("www.arcgis.com") > -1:
                urlscheme = urlparse(url).scheme
                return "{scheme}://www.arcgis.com/sharing/rest/".format(scheme=urlscheme)
            elif url.lower().endswith("sharing/"):
                return url + 'rest/'
            elif url.lower().endswith("sharing/rest/"):
                return url
            else:
                return url + 'sharing/rest/'
        return url
    #----------------------------------------------------------------------
    @property
    def product(self):
        if self._product is None:
            self._product = self._check_product()
        return self._product
    #----------------------------------------------------------------------
    @property
    def connection(self):
        """gets/sets an additional connection object to get a token from"""
        return self._connection
    #----------------------------------------------------------------------
    @connection.setter
    def connection(self, value):
        """gets/sets an additional connection object to get a token from"""
        if self._connection != value:
            self._connection = value
            self._token = None
            self._server_token = None
    #----------------------------------------------------------------------
    @property
    def token(self):
        """gets/sets the token"""
        if self.connection and self.service_url:
            return self.connection.generate_portal_server_token(serverUrl=self.service_url)
        elif self._connection and self._server_token is None:
            #create a portalserver token
            if self._connection.product == "AGO":
                return self.connection.token
            return self.generate_portal_server_token(serverUrl=self.baseurl)
        elif self._connection and self._server_token:
            return self._server_token
        elif self._connection is None and self.product == "FEDERATED_SERVER":
            self._connection = _ArcGISConnection(baseurl=self.baseurl, connection=self)
            return self.token
        elif self._token:
            return self._token
        elif self._token is None and \
             self._is_arcpy:
            return self.login(username=None, password=None)
        elif self._username and self._password:
            self.login(username=self._username, password=self._password, expiration=60)
            return self._token
        return None
    #----------------------------------------------------------------------
    @property
    def service_url(self):
        """gets/sets the service url"""
        return self._service_url
    #----------------------------------------------------------------------
    @service_url.setter
    def service_url(self, value):
        """gets/sets the service url"""
        if value:
            self._service_url = value
        else:
            self._service_url = None
    #----------------------------------------------------------------------
    @token.setter
    def token(self, value):
        """gets/sets the token"""
        if self._token != value:
            self._token = value
    #----------------------------------------------------------------------
    def generate_token(self, username, password, expiration=60, client_id=None):
        """ Generates and returns a new token, but doesn't re-login. """
        if self._is_arcpy and \
           self.product in ("PORTAL", "AGO"):
            try:
                import arcpy
                resp = arcpy.GetSigninToken()
                if 'referer' in resp:
                    self._referer = resp['referer']
                if 'token' in resp:
                    return resp['token']
                else:
                    raise arcpy.ExecuteError("Could not login using Pro Authentication")
            except ImportError as ie:
                raise Error("Could not import arcpy")
            except:
                raise arcpy.ExecuteError("Could not login using Pro Authentication")
        if self.product == "SERVER":
            postdata = { 'username': username,
                         'password': password,
                         'client': 'requestip',
                         'expiration': expiration,
                         'f': 'json' }
        else:
            postdata = { 'username': username, 'password': password,
                         'client': 'referer', 'referer': self._referer,
                         'expiration': expiration, 'f': 'json' }
        if client_id is not None:
            return self.oauth_authenticate(client_id, expiration)

        else:
            if self._tokenurl is None:
                if self.baseurl.endswith('/'):
                    resp = self.post('generateToken', postdata,
                                     ssl=True, add_token=False)
                else:
                    resp = self.post('/generateToken', postdata,
                                     ssl=True, add_token=False)
            else:
                resp = self.post(path=self._tokenurl, postdata=postdata,
                                 ssl=True, add_token=False)
            if resp:
                return resp.get('token')

    def oauth_authenticate(self, client_id, expiration):

        parameters = {
            'client_id': client_id,
            'response_type': 'code',
            'expiration': -1, # we want refresh_token to work for the life of the script
            'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob'
        }

        code = None

        if self._username is not None and self._password is not None: # built-in user through OAUTH
            content = self.get('oauth2/authorize', parameters, ssl=True, try_json=False, add_token=False)
            import re
            import json
            from bs4 import BeautifulSoup
            pattern = re.compile('var oAuthInfo = ({.*?});', re.DOTALL)
            soup = BeautifulSoup(content, 'html.parser')
            for script in soup.find_all('script'):
                script_code = str(script.string).strip()
                matches = pattern.search(script_code)
                if not matches is None:
                    js_object = matches.groups()[0]
                    oauth_info = json.loads(js_object)
                    break

            parameters = {
                'user_orgkey': '',
                'username': self._username,
                'password': self._password,
                'oauth_state': oauth_info['oauth_state']
            }
            content = self.post('oauth2/signin', parameters, ssl=True, try_json=False, add_token=False)
            soup = BeautifulSoup(content, 'html.parser')

            if soup.title is not None:
                if 'SUCCESS' in soup.title.string:
                    code = soup.title.string[len('SUCCESS code='):]

        if code is None: # try interactive signin
            url = self.baseurl + 'oauth2/authorize'
            paramstring = urlencode(parameters)
            codeurl = "{}?{}".format(url, paramstring)

            import webbrowser
            import getpass

            print("Please sign in to your GIS and paste the code that is obtained below.")
            print("If a web browser does not automatically open, please navigate to the URL below yourself instead.")
            print("Opening web browser to navigate to: " + codeurl)
            webbrowser.open_new(codeurl)
            code = getpass.getpass("Enter code obtained on signing in using SAML: ")

        if code is not None:
            parameters = {
                'client_id': client_id,
                'grant_type': 'authorization_code',
                'code': code,
                'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob'
            }
            token_info = self.post('oauth2/token', parameters, ssl=True, add_token=False)
            # print('******' + str(token_info))

            self._refresh_token = token_info['refresh_token']
            self._token = token_info['access_token']

            return self._token
        else:
            print("Unable to sign in using OAUTH")
            return None

    #----------------------------------------------------------------------
    def generate_portal_server_token(self, serverUrl, expiration=1440):
        """generates a server token using Portal token"""
        if self._auth.lower() == "pki":
            from urllib.request import HTTPCookieProcessor
            token = ""
            cj = None
            for handler in self._handlers:
                if isinstance(handler, (HTTPCookieProcessor)):
                    from urllib.parse import unquote
                    for cookie in handler.cookiejar:
                        if cookie.name.lower() == "esri_auth":
                            auth = json.loads(unquote(cookie.value))
                            if 'token' in auth:
                                token = auth['token']
                                break
                        del cookie
                        #print( [cookie.name, cookie.value, cookie.domain])

        else:
            token = self.token
        postdata = {'serverURL':serverUrl,
                    'token': token,
                    'expiration':str(expiration),
                    'f': 'json',
                    'request':'getToken',
                    'referer':self._referer}
        if self._tokenurl is None:
            if self.baseurl.endswith('/'):
                resp = self.post('generateToken', postdata,
                                 ssl=True, add_token=False)
            else:
                resp = self.post('/generateToken', postdata,
                                 ssl=True, add_token=False)
        else:
            resp = self.post(path=self._tokenurl, postdata=postdata,
                             ssl=True, add_token=False)
        if resp:
            return resp.get('token')
    #----------------------------------------------------------------------
    def login(self, username, password, expiration=60, client_id=None):
        """ Logs into the portal using username/password. """
        newtoken = None
        try:
            if self._is_arcpy: # PRO authentication
                newtoken = self.generate_token(username, password, expiration)
                if newtoken:
                    self._token = newtoken

            resp = self.post('', { 'f': 'json' }, add_token=False) # probe portal to find auth scheme
                                                  # if basic, digest, NTLM or Kerberos, etc is being used
                                                  # except handler will catch it and set self._auth appropriately

            if username is not None and password is not None:
                newtoken = self.generate_token(username, password, expiration, client_id)

                if newtoken:
                    self._token = newtoken
                    self._username = username
                    self._password = password
                    self._expiration = expiration

                    return newtoken

            elif client_id is not None:
                newtoken = self.generate_token(username, password, expiration, client_id)
                return newtoken

            elif self._is_arcpy:
                return newtoken

            else:
                self._auth = "ANON"

        except HTTPError as err:
            if err.code == 401:
                authhdr = err.headers.get('WWW-Authenticate')
                if authhdr is not None:
                    if authhdr.lower().startswith('basic'):
                        self._auth = "BASIC"
                    elif authhdr.lower().startswith('digest'):
                        self._auth = "DIGEST"
                    elif authhdr.lower().startswith('ntlm'):
                        self._auth = "IWA"
                    elif authhdr.lower().startswith('negotiate'):
                        self._auth = "IWA"
                    else:
                        _log.warning('Unsupported authentication scheme: ' + authhdr)

                    return newtoken

            else:
                raise
        except ValueError as ve:
            if str(ve) == "AbstractBasicAuthHandler does not support the following scheme: 'Negotiate'":
                self._auth = "IWA"


    #----------------------------------------------------------------------
    def relogin(self, expiration=60):
        """ Re-authenticates with the portal using the same username/password. """
        if self._refresh_token is not None and self._client_id is not None: # oauth2
            parameters = {
                'client_id': self._client_id,
                'grant_type': 'refresh_token',
                'refresh_token': self._refresh_token,
                'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob'
            }
            token_info = self.post('oauth2/token', parameters, ssl=True, add_token=False)
            if 'refresh_token' in token_info:
                self._refresh_token = token_info['refresh_token']
            self._token = token_info['access_token']

            return self._token
        else:
            return self.login(self._username, self._password, expiration)
    #----------------------------------------------------------------------
    def logout(self):
        """ Logs out of the portal. """
        self._token = None
        self._server_token = None
    #----------------------------------------------------------------------
    @property
    def is_logged_in(self):
        """ Returns true if logged into the portal. """
        return self._token is not None or self._server_token is not None
    #----------------------------------------------------------------------
    def _mainType(self, resp):
        """ gets the main type from the response object"""
        if six.PY2:
            return resp.headers.maintype
        elif six.PY3:
            return resp.headers.get_content_maintype()
        else:
            return None
    #----------------------------------------------------------------------
    def _check_product(self):
        """
        determines if the product is portal, arcgis online or arcgis server
        """
        baseurl = self.baseurl
        if baseurl is None:
            return "UNKNOWN"
        if baseurl.lower().find("arcgis.com") > -1:
            return "AGO"
        elif baseurl.lower().find("/sharing/rest") > -1:
            return "PORTAL"
        else:
            #Brute Force Method
            root = baseurl.lower().split("/sharing")[0]
            root = baseurl.lower().split('/rest')[0]
            parts = ['/info', '/rest/info', '/sharing/rest/info']
            params = {"f" : "json"}
            for pt in parts:
                try:
                    res = self.get(path=root + pt, params=params)
                    if self._tokenurl is None and \
                       res is not None and \
                       'authInfo' in res and \
                       'tokenServicesUrl' in res['authInfo']:
                        self._tokenurl = res['authInfo']['tokenServicesUrl']
                except HTTPError as e:
                    res = ""
                if isinstance(res, dict) and \
                   "currentVersion" in res:
                    t_parsed = urlparse(self._tokenurl[1:]).path
                    b_parsed = urlparse(self.baseurl[1:]).path
                    if t_parsed.startswith("/"):
                        t_parsed = t_parsed[1:].split("/")[0]
                    else:
                        t_parsed = t_parsed.split("/")[0]
                    if b_parsed.startswith("/"):
                        b_parsed = b_parsed[1:].split("/")[0]
                    else:
                        b_parsed = b_parsed.split("/")[0]
                    if t_parsed.lower() != b_parsed.lower():
                        return "FEDERATED_SERVER"
                    return "SERVER"
                del pt
                del res
        return "PORTAL"
    #----------------------------------------------------------------------
    def _get_file_name(self, contentDisposition,
                       url, ext=".unknown"):
        """ gets the file name from the header or url if possible """
        if six.PY2:
            if contentDisposition is not None:
                return re.findall(r'filename[^;=\n]*=(([\'"]).*?\2|[^;\n]*)',
                                  contentDisposition.strip().replace('"', ''))[0][0]
            elif os.path.basename(url).find('.') > -1:
                return os.path.basename(url)
        elif six.PY3:
            if contentDisposition is not None:
                p = re.compile(r'filename[^;=\n]*=(([\'"]).*?\2|[^;\n]*)')
                return p.findall(contentDisposition.strip().replace('"', ''))[0][0]
            elif os.path.basename(url).find('.') > -1:
                return os.path.basename(url)

        if six.PY2:
            hex = '-===============%s==' % uuid.uuid4().get_hex()
        elif six.PY3:
            hex = '-===============%s==' % uuid.uuid4().hex

        return "%s.%s" % (hex, ext)
    #----------------------------------------------------------------------
    def _process_response(self, resp, out_folder=None,  file_name=None, force_bytes=False):
        """ processes the response object"""
        CHUNK = 4056
        maintype = self._mainType(resp)
        contentDisposition = resp.headers.get('content-disposition')
        contentType = resp.headers.get('content-type')
        contentLength = resp.headers.get('content-length')
        if not force_bytes and \
           (maintype.lower() in ('image', 'application/x-zip-compressed') or \
            contentType == 'application/x-zip-compressed' or \
           (contentDisposition is not None and contentDisposition.lower().find('attachment;') > -1)) or \
           (contentDisposition is not None and contentDisposition.find("filename=") > -1):
            fname = self._get_file_name(contentDisposition=contentDisposition, url=resp.geturl()).split('?')[0]
            if out_folder is None:
                out_folder = tempfile.gettempdir()
            if contentLength is not None:
                max_length = int(contentLength)
                if max_length < CHUNK:
                    CHUNK = max_length
            if file_name is None:
                file_name = os.path.join(out_folder, fname)
            else:
                file_name = os.path.join(out_folder, file_name)
            with open(file_name, 'wb') as writer:
                for data in self._chunk(response=resp):
                    writer.write(data)
                    del data
                del writer
            return file_name, True
        else:
            read = ""
            if file_name and out_folder:
                f_n_path = os.path.join(out_folder, file_name)
                with open(f_n_path, 'wb') as writer:
                    for data in self._chunk(response=resp, size=4096):
                        writer.write(data)
                        del data
                    writer.flush()
                    del writer
                return f_n_path, True
            else:
                for data in self._chunk(response=resp, size=4096):
                    if six.PY3 == True:
                        if read == "":
                            read = data
                        else:
                            read += data
                    else:
                        read += data
                    del data
            if six.PY3 and len(read) > 0:
                try:
                    read = read.decode("utf-8").strip()
                except:
                    pass
            try:
                return read.strip(), False
            except:
                return read, False
        return "", False
    #----------------------------------------------------------------------
    def _chunk(self, response, size=4096):
        """
        downloads a web response in pieces to ensure there are no
        memory issues.
        """
        method = response.headers.get("content-encoding")
        if method == "gzip":
            d = zlib.decompressobj(16+zlib.MAX_WBITS)
            b = response.read(size)
            while b:
                data = d.decompress(b)
                yield data
                b = response.read(size)
                del data
        else:
            while True:
                chunk = response.read(size)
                if not chunk: break
                yield chunk
    #----------------------------------------------------------------------
    def get(self, path, params=None, ssl=False,
            compress=True, try_json=True, is_retry=False,
            use_ordered_dict=False, out_folder=None,
            file_name=None, force_bytes=False, add_token=True,
            token=DEFAULT_TOKEN):
        """ Returns result of an HTTP GET. Handles token timeout and all SSL mode."""
        url = path
        if url.lower().find("https://") > -1 or\
           url.lower().find("http://") > -1:
            url = path
        elif len(url) == 0:
            url = self.baseurl
        elif (len(url) > 0 and url[0] == '/' ) == False and \
             self.baseurl.endswith('/') == False:
            url = "/{path}".format(path=url)

        if not url.startswith('http://') and \
           not url.startswith('https://'):
            url = self.baseurl + url
        if ssl or self.all_ssl:
            url = url.replace('http://', 'https://')

        if params is None:
            params = {}
        if try_json:
            params['f'] = 'json'


        if add_token:
            if token != DEFAULT_TOKEN: # use the provided token, if any
                if token is not None:
                    params['token'] = token
                else:
                    pass # no token, public access
            elif self.token is not None:
                params['token'] = self.token

        if len(params.keys()) > 0:
            params = {k: jsonize_dict(v) for k, v in params.items()}
            url = "{url}?{params}".format(url=url,
                                          params=urlencode(params))
        _log.debug('REQUEST (get): ' + url)

        try:
            # Send the request and read the response
            if self._auth.lower() == "pki":
                if self._useragent is None:
                    self._useragent = 'geosaurus/' + __version__
                headers = [('User-Agent', self._useragent)]
            else:
                headers = [('Referer', self._referer),
                           ('User-Agent', self._useragent)]
            if compress:
                headers.append(('Accept-encoding', 'gzip'))
            if self._handlers is None or self._auth in ('BASIC', 'DIGEST', 'IWA'):
                self._handlers = self.get_handlers(verify_cert=self._verify_cert)
            opener = request.build_opener(*self._handlers)
            opener.addheaders = headers
            resp = opener.open(url)

            resp_data, is_file = self._process_response(resp,
                                               out_folder=out_folder,
                                               file_name=file_name,
                                               force_bytes=force_bytes)

            # If is a file or we're not trying to parse to JSON, return response as is
            if is_file or not try_json:
                return resp_data

            try:
                if use_ordered_dict:
                    resp_json = json.loads(resp_data,
                                           object_pairs_hook=OrderedDict)
                else:
                    resp_json = json.loads(resp_data)

                # Check for errors, and handle the case where the token timed
                # out during use (and simply needs to be re-generated)
                try:
                    if resp_json:
                        if 'error' in resp_json:
                            errorcode = resp_json['error']['code'] if 'code' in resp_json['error'] else 0
                            if errorcode == 498 and not is_retry:
                                _log.info('Token expired during get request, ' \
                                          + 'fetching a new token and retrying')
                                newtoken = self.relogin()


                                self.token = newtoken

                                if token != DEFAULT_TOKEN: # was provided a FEDERATED SERVER token, that has expired
                                    newtoken = self.generate_portal_server_token(url)
                                else:
                                    newtoken = newtoken

                                newpath = self._url_add_token(path, newtoken)


                                return self.get(path=newpath, params=params, ssl=ssl, compress=compress, try_json=try_json, is_retry=True)
                            elif errorcode == 498:
                                raise RuntimeError('Invalid token')

                            elif errorcode == 403:
                                message = resp_json['error']['message'] if 'message' in resp_json['error'] else ''
                                if message == "SSL Required":
                                    return self.get(path=path, params=params, ssl=True, compress=compress, try_json=try_json, is_retry=True)

                            self._handle_json_error(resp_json['error'], errorcode)
                            return None
                except AttributeError:
                    # Top-level JSON object isnt a dict, so can't have an error
                    pass

                # If the JSON parsed correctly and there are no errors,
                # return the JSON
                return resp_json

            # If we couldnt parse the response to JSON, return it as is
            except ValueError:
                return resp_data
            except TypeError as te:
                _log.info(te.args[0])
                return resp_data

        # If we got an HTTPError when making the request check to see if it's
        # related to token timeout, in which case, regenerate a token
        except HTTPError as e:
            if e.code == 498 and not is_retry:
                _log.info('Token expired during get request, fetching a new ' \
                          + 'token and retrying')
                self.logout()
                newtoken = self.relogin()
                newpath = self._url_add_token(path, newtoken)
                return self.get(newpath, ssl, try_json, is_retry=True)
            elif e.code == 498:
                raise RuntimeError('Invalid token')
            else:
                raise e
    #----------------------------------------------------------------------
    def _ensure_dir(self, f):
        if not os.path.exists(f):
            os.makedirs(f)
    #----------------------------------------------------------------------
    def _url_add_token(self, url, token):

        # Parse the URL and query string
        urlparts = urlparse(url)
        qs_list = parse_qsl(urlparts.query)

        # Update the token query string parameter
        replaced_token = False
        new_qs_list = []
        for qs_param in qs_list:
            if qs_param[0] == 'token':
                qs_param = ('token', token)
                replaced_token = True
            new_qs_list.append(qs_param)
        if not replaced_token:
            new_qs_list.append(('token', token))

        # Rebuild the URL from parts and return it
        return urlunparse((urlparts.scheme, urlparts.netloc,
                           urlparts.path, urlparts.params,
                           urlencode(new_qs_list),
                           urlparts.fragment))
    #----------------------------------------------------------------------
    def get_handlers(self, verify_cert=True):
        from urllib.request import HTTPRedirectHandler
        redirect_handler = HTTPRedirectHandler()
        redirect_handler.max_redirections = 30
        redirect_handler.max_repeats = 30
        handlers = [redirect_handler]

        if self.proxy_host: # simple proxy support
            from urllib.request import ProxyHandler
            if self.proxy_port is None:
                self.proxy_port = 80
            proxies = {"http":"http://%s:%s" % (self.proxy_host, self.proxy_port),
                       "https":"https://%s:%s" % (self.proxy_host, self.proxy_port)}
            proxy_support = ProxyHandler(proxies)
            handlers.append(proxy_support)

        if self._auth == "BASIC": # used by LDAP
            passman = request.HTTPPasswordMgrWithDefaultRealm()
            passman.add_password(None,
                                 self._parsed_org_url,
                                 self._username,
                                 self._password)
            handlers.append(request.HTTPBasicAuthHandler(passman))

        if self._auth == "DIGEST":
            passman = request.HTTPPasswordMgrWithDefaultRealm()
            passman.add_password(None,
                                 self._parsed_org_url,
                                 self._username,
                                 self._password)
            handlers.append(request.HTTPDigestAuthHandler(passman))

        elif self._auth == "IWA":
            if os.name == 'nt':
                try:
                    from .common._iwa import NtlmSspiAuthHandler, KerberosSspiAuthHandler

                    try:
                        auth_NTLM = NtlmSspiAuthHandler()
                        handlers.append(auth_NTLM)
                    except:
                        pass

                    auth_krb = KerberosSspiAuthHandler()
                    handlers.append(auth_krb)

                except Error as err:
                    _log.error("winkerberos packages is required for IWA authentication (NTLM and Kerberos).")
                    _log.error("Please install it:\n\tconda install winkerberos")
                    _log.error(str(err))
            else:
                _log.error('The GIS uses Integrated Windows Authentication which is currently only supported on the Windows platform')

        elif self._auth == "PKI":
            handlers.append(HTTPSClientAuthHandler(self.key_file, self.cert_file))

        cj = cookiejar.CookieJar()
        handlers.append(request.HTTPCookieProcessor(cj))



        if not verify_cert or not self._verify_cert:
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
            handler = request.HTTPSHandler(context=ctx)
            handlers.append(handler)

        return handlers
    #----------------------------------------------------------------------

    def post(self, path, postdata=None, files=None, ssl=False, compress=True,
             is_retry=False, use_ordered_dict=False, add_token=True, verify_cert=True,
             token=DEFAULT_TOKEN, try_json=True, out_folder=None,
             file_name=None, force_bytes=False, add_headers=None):
        """ Returns result of an HTTP POST. Supports Multipart requests."""
        # prevent double encoding
        if not is_retry:
            path = quote(path, ':/')
        url = path
        if url.lower().find("https://") > -1 or\
           url.lower().find("http://") > -1:
            url = path
        elif len(url) == 0:
            url = self.baseurl
        elif (len(url) > 0 and url[0] == '/' ) == False and \
           self.baseurl.endswith('/') == False:
            url = "/{path}".format(path=url)

        if not url.startswith('http://') and \
           not url.startswith('https://'):
            url = self.baseurl + url

        if ssl or self.all_ssl:
            url = url.replace('http://', 'https://')
        #if verify_cert == False:
        #    import ssl
        #    ssl._create_default_https_context = ssl._create_unverified_context
        # Add the token if logged in
        if add_token:
            if token != DEFAULT_TOKEN: # use the provided token, if any
                if token is not None:
                    postdata['token'] = token
                else:
                    pass # no token, public access
            elif self.token is not None:
                postdata['token'] = self.token

        if _log.isEnabledFor(logging.DEBUG):
            msg = 'REQUEST: ' + url + ', ' + str(postdata)
            if files:
                msg += ', files=' + str(files)
            _log.debug(msg)

        # If there are files present, send a multipart request
        if files is not None:
            #parsed_url = urlparse(url)
            mpf = MultiPartForm(param_dict=postdata, files=files)
            req = request.Request(url)
            body = mpf.make_result
            req.add_header('User-agent', self._useragent)
            req.add_header('Content-type', mpf.get_content_type())
            req.add_header('Content-length', len(body))
            if isinstance(add_headers, list):
                for ah in add_headers:
                    req.add_header(ah[0], ah[1])
            req.data = body
            headers = [('User-Agent', self._useragent),
                       ('Content-type', mpf.get_content_type()),
                       ('Content-length', len(body))]
            if self._referer:
                headers.append(('Referer', self._referer))
            if isinstance(add_headers, list):
                for ah in add_headers:
                    headers.append(ah)
            if compress:
                headers.append(('Accept-encoding', 'gzip'))
            if self._handlers is None or self._auth in ('BASIC', 'DIGEST', 'IWA'):
                self._handlers = self.get_handlers(verify_cert)
            opener = request.build_opener(*self._handlers)

            opener.addheaders = headers

            resp = opener.open(req)
            resp_data, is_file = self._process_response(resp,
                                               out_folder=out_folder,
                                               file_name=file_name,
                                               force_bytes=force_bytes)
        # Otherwise send a normal HTTP POST request
        else:
            encoded_postdata = None
            if postdata:
                postdata = {k: jsonize_dict(v) for k, v in postdata.items()}
                encoded_postdata = urlencode(postdata)
            if self._referer:
                headers = [('Referer', self._referer),
                           ('User-Agent', self._useragent)]
            else:
                headers = [('User-Agent', self._useragent)]
            if compress:
                headers.append(('Accept-encoding', 'gzip'))

            if self._handlers is None or self._auth in ('BASIC', 'DIGEST', 'IWA'):
                self._handlers = self.get_handlers(verify_cert)
            opener = request.build_opener(*self._handlers)

            opener.addheaders = headers
            #print("***"+url)
            resp = opener.open(url, data=encoded_postdata.encode())
            resp_data, is_file = self._process_response(resp,
                                               out_folder=out_folder,
                                               file_name=file_name,
                                               force_bytes=force_bytes)

        # Parse the response into JSON
        if _log.isEnabledFor(logging.DEBUG):
            _log.debug('RESPONSE: ' + url + ', ' + resp_data)
        # print(resp_data)

        # If is a file or we're not trying to parse to JSON, return response as is
        if is_file or not try_json:
            return resp_data

        if use_ordered_dict:
            resp_json = json.loads(resp_data, object_pairs_hook=OrderedDict)
        else:
            resp_json = json.loads(resp_data)


        # Check for errors, and handle the case where the token timed out
        # during use (and simply needs to be re-generated)
        try:
            if resp_json.get('error', None):

                errorcode = resp_json['error']['code'] if 'code' in resp_json['error'] else 0
                if errorcode == 498 and not is_retry:
                    _log.info('Token expired during post request, fetching a new '
                              + 'token and retrying')
                    self.logout()
                    newtoken = self.relogin()

                    self.token = newtoken
                    retry_token = None
                    if token != DEFAULT_TOKEN: # was provided a token, that has expired
                        newfedtoken = self.generate_portal_server_token(url)
                        retry_token = newfedtoken
                    else:
                        retry_token = newtoken

                    return self.post(path, postdata, files, ssl, compress, token=retry_token, verify_cert=verify_cert,
                                     is_retry=True)
                elif errorcode == 498:
                    raise RuntimeError('Invalid token')

                elif errorcode == 403:
                    message = resp_json['error']['message'] if 'message' in resp_json['error'] else ''
                    if message == "SSL Required":
                        return self.post(path, postdata, files, ssl=True, compress=compress, token=token,
                                         verify_cert=verify_cert, is_retry=True)

                self._handle_json_error(resp_json['error'], errorcode)
                return None

        except AttributeError:
            # Top-level JSON object isnt a dict, so can't have an error
            pass

        return resp_json
    #----------------------------------------------------------------------
    def _get_content_type(self, filename):
        return mimetypes.guess_type(filename)[0] or 'application/octet-stream'
    #----------------------------------------------------------------------
    def _handle_json_error(self, error, errorcode):
        errormessage = error.get('message', 'Unknown Error')
        _log.error(errormessage)
        if 'details' in error and error['details'] is not None:
            for errordetail in error['details']:
                errormessage = errormessage + "\n" + errordetail
                _log.error(errordetail)

        errormessage = errormessage + "\n(Error Code: " + str(errorcode) +")"
        raise RuntimeError(errormessage)

class _StrictURLopener(request.FancyURLopener):
    def http_error_default(self, url, fp, errcode, errmsg, headers):
        if errcode != 200:
            raise HTTPError(url, errcode, errmsg, headers, fp)

def _normalize_url(url, charset='utf-8'):
    """ Normalizes a URL. Based on http://code.google.com/p/url-normalize."""
    def _clean(string):
        string = str(unquote(string), 'utf-8', 'replace')
        return unicodedata.normalize('NFC', string).encode('utf-8')

    default_port = {
        'ftp': 21,
        'telnet': 23,
        'http': 80,
        'gopher': 70,
        'news': 119,
        'nntp': 119,
        'prospero': 191,
        'https': 443,
        'snews': 563,
        'snntp': 563,
    }

    # if there is no scheme use http as default scheme
    if url[0] not in ['/', '-'] and ':' not in url[:7]:
        url = 'http://' + url


    # shebang urls support
    url = url.replace('#!', '?_escaped_fragment_=')

    # splitting url to useful parts
    scheme, auth, path, query, fragment = urlsplit(url.strip())
    (userinfo, host, port) = re.search('([^@]*@)?([^:]*):?(.*)', auth).groups()

    # Always provide the URI scheme in lowercase characters.
    scheme = scheme.lower()

    # Always provide the host, if any, in lowercase characters.
    host = host.lower()
    if host and host[-1] == '.':
        host = host[:-1]
    # take care about IDN domains
    host = host.decode(charset).encode('idna')  # IDN -> ACE

    # Only perform percent-encoding where it is essential.
    # Always use uppercase A-through-F characters when percent-encoding.
    # All portions of the URI must be utf-8 encoded NFC from Unicode strings
    path = quote(_clean(path), "~:/?#[]@!$&'()*+,;=")
    fragment = quote(_clean(fragment), "~")

    # note care must be taken to only encode & and = characters as values
    query = "&".join(["=".join([quote(_clean(t), "~:/?#[]@!$'()*+,;=") \
                                for t in q.split("=", 1)]) for q in query.split("&")])

    # Prevent dot-segments appearing in non-relative URI paths.
    if scheme in ["", "http", "https", "ftp", "file"]:
        output = []
        for part in path.split('/'):
            if part == "":
                if not output:
                    output.append(part)
            elif part == ".":
                pass
            elif part == "..":
                if len(output) > 1:
                    output.pop()
            else:
                output.append(part)
        if part in ["", ".", ".."]:
            output.append("")
        path = '/'.join(output)

    # For schemes that define a default authority, use an empty authority if
    # the default is desired.
    if userinfo in ["@", ":@"]:
        userinfo = ""

    # For schemes that define an empty path to be equivalent to a path of "/",
    # use "/".
    if path == "" and scheme in ["http", "https", "ftp", "file"]:
        path = "/"

    # For schemes that define a port, use an empty port if the default is
    # desired
    if port and scheme in list(default_port.keys()):
        if port.isdigit():
            port = str(int(port))
            if int(port) == default_port[scheme]:
                port = ''

    # Put it all back together again
    auth = (userinfo or "") + host
    if port:
        auth += ":" + port
    if url.endswith("#") and query == "" and fragment == "":
        path += "#"
    return urlunsplit((scheme, auth, path, query, fragment))

def _parse_hostname(url, include_port=False):
    """ Parses the hostname out of a URL."""
    if url:
        parsed_url = urlparse((url))
        return parsed_url.netloc if include_port else parsed_url.hostname

def _is_http_url(url):
    if url:
        return urlparse(url).scheme in ['http', 'https']

def _unpack(obj_or_seq, key=None, flatten=False):
    """ Turns a list of single item dicts in a list of the dict's values."""

    # The trivial case (passed in None, return None)
    if not obj_or_seq:
        return None

    # We assume it's a sequence
    new_list = []
    for obj in obj_or_seq:
        value = _unpack_obj(obj, key, flatten)
        new_list.extend(value)

    return new_list

def _unpack_obj(obj, key=None, flatten=False):
    try:
        if key:
            value = [obj.get(key)]
        else:
            value = list(obj.values())
    except AttributeError:
        value = [obj]

    # Flatten any lists if directed to do so
    if value and flatten:
        value = [item for sublist in value for item in sublist]

    return value

def _remove_non_ascii(s):
    return ''.join(i for i in s if ord(i) < 128)

def _tostr(obj):
    if not obj:
        return ''
    if isinstance(obj, list):
        return ', '.join(map(_tostr, obj))
    return str(obj)




# This function is a workaround to deal with what's typically described as a
# problem with the web server closing a connection. This is problem
# experienced with www.arcgis.com (first encountered 12/13/2012). The problem
# and workaround is described here:
# http://bobrochel.blogspot.com/2010/11/bad-servers-chunked-encoding-and.html
def _patch_http_response_read(func):
    def inner(*args):
        try:
            return func(*args)
        except http_client.IncompleteRead as e:
            return e.partial

    return inner
http_client.HTTPResponse.read = _patch_http_response_read(http_client.HTTPResponse.read)



"Network analysis tools."

from ._vrp import solve_vehicle_routing_problem
from ._closest_facility import find_closest_facilities
from ._location_allocation import solve_location_allocation
from ._od import generate_origin_destination_cost_matrix
from ._route import find_routes
from ._routing_utils import get_travel_modes
from ._svcareas import generate_service_areas



"""
These functions help you identify, quantify, and visualize spatial patterns in your data.

calculate_density takes known quantities of some phenomenon and spreads these quantities across the map.
find_hot_spots identifies statistically significant clustering in the spatial pattern of your data.
interpolate_points predicts values at new locations based on measurements found in a collection of points.
"""

import arcgis as _arcgis

def calculate_density(
        input_layer,
        field=None,
        cell_size=None,
        cell_size_units="Meters",
        radius=None,
        radius_units=None,
        bounding_polygon_layer=None,
        area_units=None,
        classification_type="EqualInterval",
        num_classes=10,
        output_name=None,
        context=None,
        gis=None):
    """
    The calculate_density function creates a density map from point or line features by spreading known quantities of
    some phenomenon (represented as attributes of the points or lines) across the map. The result is a layer of areas
    classified from least dense to most dense.

    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)
        The point or line features from which to calculate density.
    field : Optional string
        A numeric field name specifying the number of incidents at each location. If not specified, each location will
        be assumed to represent a single count.
    cell_size : Optional float
        This value is used to create a mesh of points where density values are calculated. The default is approximately
        1/1000th of the smaller of the width and height of the analysis extent as defined in the context parameter.
    cell_size_units : Optional string
        The units of the cellSize value
    radius : Optional float
        A distance specifying how far to search to find point or line features when calculating density values.
    radius_units : Optional string
        The units of the radius parameter.
    bounding_polygon_layer : Optional layer (see Feature Input in documentation)
        A layer specifying the polygon(s) where you want densities to be calculated.
    area_units : Optional string
        The units of the calculated density values.
    classification_type : Optional string
        Determines how density values will be classified into polygons.
    num_classes : Optional int
        This value is used to divide the range of predicted values into distinct classes. The range of values in each
        class is determined by the classificationType parameter.
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    result_layer : layer (FeatureCollection)
    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.calculate_density(
        input_layer,
        field,
        cell_size,
        cell_size_units,
        radius,
        radius_units,
        bounding_polygon_layer,
        area_units,
        classification_type,
        num_classes,
        output_name,
        context)

def summarize_center_and_dispersion(
        analysis_layer,
        summarize_type,
        ellipse_size=None,
        weight_field=None,
        group_field=None,
        output_name=None,
        context=None,
        gis=None):

    """
    The Summarize Center and Dispersion task finds central features and directional distributions.

    ====================    =========================================================
    **Argument**            **Description**
    --------------------    ---------------------------------------------------------
    analysis_layer          The point, line, or polygon features to be analyzed. This
                            parameter can be a URL to a feature service layer with an
                            optional filter to select specific feaures, or a feature
                            collection
    --------------------    ---------------------------------------------------------
    summarize_type          The method with which to summarize the analysis_layer.
                            Choice List: 
                            ["CentralFeature", "MeanCenter", "MedianCenter", 
                            "Ellipse"]
                            Example: "CentralFeature"
    --------------------    ---------------------------------------------------------
    ellipse_size            The size of the output ellipse in standard deviations. 
                            The default ellipse size is 1. Valid choices are 1, 2, or
                            3 standard deviations. 
                            Choice List: [1, 2, 3]
                            Examples: 
                            "1"
                            [1, 2, 3]
    --------------------    ---------------------------------------------------------
    weight_field            A numeric field in the analysis_layer to be used to 
                            weight locations according to their relative importance.
    --------------------    ---------------------------------------------------------
    group_field             The field used to group features for separate directional
                            distribution calculations. The group_field can be of 
                            integer, date, or string type.
    --------------------    ---------------------------------------------------------
    output_name             Optional string. Additional properties such as output 
                            feature service name.                        
    --------------------    ---------------------------------------------------------
    context                 Optional string. Additional settings such as processing 
                            extent and output spatial reference.
    --------------------    ---------------------------------------------------------
    gis                     Optional, the GIS on which this tool runs. If not 
                            specified, the active GIS is used.                                                      
    ====================    =========================================================

    :returns: Python dictionary with the following keys:
        "central_feature_result_layer" : layer (FeatureCollection)
        "mean_feature_result_layer" : layer (FeatureCollection)
        "median_feature_result_layer" : layer (FeatureCollection)
        "ellipse_feature_result_layer" : layer (FeatureCollection)
        "process_info" : list of messages
    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.summarize_center_and_dispersion(
        analysis_layer,
        summarize_type,
        ellipse_size,
        weight_field,
        group_field,
        output_name,
        context) 

def find_point_clusters(
        analysis_layer,
        min_features_cluster,
        search_distance=None,
        search_distance_unit=None,
        output_name=None,
        context=None,
        gis=None):

    """
    The Find Point Clusters function finds clusters of point features in surrounding 
    noise based on their spatial distribution. Output is a layer containing records 
    assigned to a cluster or noise. 

    ====================    =========================================================
    **Argument**            **Description**
    --------------------    ---------------------------------------------------------
    analysis_layer          Required layer. The point feature layer for which 
                            density-based clustering will be calculated.
    --------------------    ---------------------------------------------------------
    min_features_cluster    Required integer. The minimum number of features to be 
                            considered a cluster. Any cluster with fewer features 
                            than the number provided will be considered noise.
    --------------------    ---------------------------------------------------------
    search_distance         Optional double. The maximum distance to consider. The 
                            Minimum Features per Cluster specified must be found 
                            within this distance for cluster membership. Individual 
                            clusters will be separated by at least this distance. If 
                            a feature is located further than this distance from the 
                            next closest feature in the cluster, it will not be 
                            included in the cluster.
    --------------------    ---------------------------------------------------------
    search_distance_unit    Optional string. The linear unit to be used for the 
                            search distance parameter.
    --------------------    ---------------------------------------------------------
    output_name             Optional string. Additional properties such as output 
                            feature service name.
    --------------------    ---------------------------------------------------------
    context                 Optional string. Additional settings such as processing 
                            extent and output spatial reference.
    --------------------    ---------------------------------------------------------
    gis                     Optional, the GIS on which this tool runs. If not 
                            specified, the active GIS is used.                                                      
    ====================    =========================================================

    :returns: Python dictionary with the following keys:
        "point_clusters_result_layer" : layer (FeatureCollection)
        "process_info" : list of messages
    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_point_clusters(
        analysis_layer,
        min_features_cluster,
        search_distance,
        search_distance_unit,
        output_name,
        context)    

def find_hot_spots(
        analysis_layer,
        analysis_field=None,
        divided_by_field=None,
        bounding_polygon_layer=None,
        aggregation_polygon_layer=None,
        output_name=None,
        context=None,
        gis=None):
    """
    The Find Hot Spots function finds statistically significant clusters of incident points, weighted points, or
    weighted polygons. For incident data, the analysis field (weight) is obtained by aggregation.
    Output is a hot spot map.

    Parameters
    ----------
    gis : The GIS used for running this analysis
    analysis_layer : Required layer (see Feature Input in documentation)
        The point or polygon feature layer for which hot spots will be calculated.
    analysis_field : Optional string
        The numeric field in the AnalysisLayer that will be analyzed.
    divided_by_field : Optional string

    bounding_polygon_layer : Optional layer (see Feature Input in documentation)
        When the analysis layer is points and no AnalysisField is specified, you can provide polygons features that
        define where incidents could have occurred.
    aggregation_polygon_layer : Optional layer (see Feature Input in documentation)
        When the AnalysisLayer contains points and no AnalysisField is specified, you can provide polygon features into
        which the points will be aggregated and analyzed, such as administrative units.
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    dict with the following keys:
       "hot_spots_result_layer" : layer (FeatureCollection)
       "process_info" : list of messages
    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_hot_spots(
        analysis_layer,
        analysis_field,
        divided_by_field,
        bounding_polygon_layer,
        aggregation_polygon_layer,
        output_name,
        context)


def find_outliers(analysis_layer,
                  analysis_field,
                  divided_by_field=None,
                  bounding_polygon_layer=None,
                  aggregation_polygon_layer=None,
                  permutations=None,
                  shape_type=None,
                  cell_size=None,
                  cell_units=None,
                  distance_band=None,
                  band_units=None,
                  output_name=None,
                  context=None,
                  gis=None):
    """

    The Find Outliers task analyzes point data (such as crime incidents, traffic accidents, or trees) or field values associated with points or area features (such as the number of people in each census tract or the total sales for retail stores). It finds statistically significant spatial clusters of high values and low values and statistically significant high or low spatial outliers within those clusters.

    The result map layer shows high outliers in red and low outliers in dark blue. Clusters of high values appear pink and clusters of low values appear light blue. Features that are beige are not a statistically significant outlier and not part of a statistically significant cluster; the spatial pattern associated with these features could very likely be the result of random processes and random chance.

    Parameters
    ----------
    analysis_layer : Required layer (see Feature Input in documentation)
        The point or polygon feature layer for which outliers will be calculated.
    analysis_field : Optional string
        The numeric field that will be analyzed.
    divided_by_field : Optional string, The numeric field in the analysis_layer that will be used to normalize your data.
    bounding_polygon_layer : Optional layer (see Feature Input in documentation)
        When the analysis layer is points and no analysisField is specified, you can provide polygon features that define where incidents could have occurred.
    aggregation_polygon_layer : Optional layer (see Feature Input in documentation)
        When the AnalysisLayer contains points and no AnalysisField is specified, you can provide polygon features into which the points will be aggregated and analyzed, such as administrative units.
    permutations : Permutations are used to determine how likely it would be to find the actual spatial distribution of the values you are analyzing. Choosing the number of permutations is a balance between precision and increased processing time. A lower number of permutations can be used when first exploring a problem, but it is best practice to increase the permutations to the highest number feasible for final results.

       - Speed implements 199 permutations and results in p-values with a precision of 0.01.
       - Balance implements 499 permutations and results in p-values with a precision of 0.002.
       - Precision implements 999 permutations and results in p-values with a precision of 0.001.
       Values: Speed | Balance | Precision
    shape_type : optional string, The shape of the polygon mesh the input features will be aggregated into.

      - Fishnet - The input features will be aggregated into a grid of square (fishnet) cells.
      - Hexagon - The input features will be aggregated into a grid of hexagonal cells.
    cell_size : The size of the grid cells used to aggregate your features. When aggregating into a hexagon grid, this distance is used as the height to construct the hexagon polygons.
    cell_units : The units of the cellSize value. You must provide a value if cellSize has been set.
      Values: Miles | Feet | Kilometers | Meters
    distance_band : The spatial extent of the analysis neighborhood. This value determines which features are analyzed together in order to assess local clustering.
    band_units : The units of the distanceBand value. You must provide a value if distanceBand has been set.
      Values: Miles | Feet | Kilometers | Meters
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis : The GIS used for running this analysis

    Returns
    -------
    Item it output_name is set.
    dict with the following keys:
       "find_outliers_result_layer" : layer (FeatureCollection)
       "process_info" : list of messages

        """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_outliers(analysis_layer,
                                                    analysis_field,
                                                    divided_by_field,
                                                    bounding_polygon_layer,
                                                    aggregation_polygon_layer,
                                                    permutations,
                                                    shape_type,
                                                    cell_size,
                                                    cell_units,
                                                    distance_band,
                                                    band_units,
                                                    output_name,
                                                    context)


def interpolate_points(
        input_layer,
        field,
        interpolate_option="5",
        output_prediction_error=False,
        classification_type="GeometricInterval",
        num_classes=10,
        class_breaks=[],
        bounding_polygon_layer=None,
        predict_at_point_layer=None,
        output_name=None,
        context=None,
        gis=None):
    """
    The Interpolate Points function allows you to predict values at new locations based on measurements from a
    collection of points. The function takes point data with values at each point and returns areas classified by
    predicted values.

    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)
        The point layer whose features will be interpolated.
    field : Required string
        Name of the numeric field containing the values you wish to interpolate.
    interpolate_option : Optional string
        Integer value declaring your preference for speed versus accuracy, from 1 (fastest) to 9 (most accurate). More
        accurate predictions take longer to calculate.
    output_prediction_error : Optional bool
        If True, a polygon layer of standard errors for the interpolation predictions will be returned in the
        predictionError output parameter.
    classification_type : Optional string
        Determines how predicted values will be classified into areas.
    num_classes : Optional int
        This value is used to divide the range of interpolated values into distinct classes. The range of values in each
        class is determined by the classificationType parameter. Each class defines the boundaries of the result
        polygons.
    class_breaks : Optional list of floats
        If classificationType is Manual, supply desired class break values separated by spaces. These values define the
        upper limit of each class, so the number of classes will equal the number of entered values. Areas will not be
        created for any locations with predicted values above the largest entered break value. You must enter at least
        two values and no more than 32.
    bounding_polygon_layer : Optional layer (see Feature Input in documentation)
        A layer specifying the polygon(s) where you want values to be interpolated.
    predict_at_point_layer : Optional layer (see Feature Input in documentation)
        An optional layer specifying point locations to calculate prediction values. This allows you to make predictions
        at specific locations of interest.
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    dict with the following keys:
       "result_layer" : layer (FeatureCollection)
       "prediction_error" : layer (FeatureCollection)
       "predicted_point_layer" : layer (FeatureCollection)
    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.interpolate_points(
        input_layer,
        field,
        interpolate_option,
        output_prediction_error,
        classification_type,
        num_classes,
        class_breaks,
        bounding_polygon_layer,
        predict_at_point_layer,
        output_name,
        context)



"""
   Adminstration.py allows users to control ArcGIS for Server 10.1+
   through the Administration REST API

"""
from __future__ import absolute_import
from .._common import BaseServer
from . import _machines, _clusters
from . import _data, _info
from . import _kml, _logs
from . import _security, _services
from . import _system
from . import _uploads, _usagereports
from . import _mode
from .. import  ServicesDirectory
from arcgis._impl.connection import _ArcGISConnection
from .._common import ServerConnection
########################################################################
class Server(BaseServer):
    """
    An ArcGIS Enterprise Server site used for hosting GIS Web services.


    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string. The URL to the ArcGIS Server administration
                           end point for the ArcGIS Server Site.

                           Example: https://gis.mysite.com/arcgis/admin

                           The URL should be formatted as follows:
                           <scheme>://<fully_qualified_domain_name>:<port (optional)>/<web adaptor>/admin

                           Note: Using the fully-qualified domain name to the server, also known as the
                           Web Context URL, is recommended as generally the SSL Certificate binding for
                           the web server uses this hostname.
    ------------------     --------------------------------------------------------------------
    gis                    Optional string. The GIS object representing the Portal which thi
                           Server is federated with. The GIS object should be logged in with a username
                           in the publisher or administrator Role in order to administer the Server
    ==================     ====================================================================

    =====================     ====================================================================
    **Optional Argument**     **Description**
    ---------------------     --------------------------------------------------------------------
    baseurl                   Optional string. The root URL to a site.
                              Example: https://mysite.com/arcgis
    ---------------------     --------------------------------------------------------------------
    tokenurl                  Optional string. Used when a site is federated or when the token
                              URL differs from the site's baseurl.  If a site is federated, the
                              token URL will return as the Portal token and ArcGIS Server users
                              will not validate correctly.
    ---------------------     --------------------------------------------------------------------
    username                  Optional string. The login username for BUILT-IN GIS Server security.
    ---------------------     --------------------------------------------------------------------
    password                  Optional string. A secret word or phrase that must be used to gain
                              access to the account above.
    ---------------------     --------------------------------------------------------------------
    key_file                  Optional string. The path to a PKI key file used to authenticate the
                              user to the Web Server in front of the ArcGIS Server site.
    ---------------------     --------------------------------------------------------------------
    cert_file                 Optional string. The path to PKI cert file used to authenticate the
                              user to the Web Server in front of the ArcGIS Server site.
    ---------------------     --------------------------------------------------------------------
    proxy_host                Optional string. The web address to the proxy host if the environment
                              where the Python API is running requires a proxy host for access to the
                              Site URL or GIS URL.

                              Example: proxy.mysite.com
    ---------------------     --------------------------------------------------------------------
    proxy_port                Optional integer. The port which the proxy is accessed through,
                              default is 80.
    ---------------------     --------------------------------------------------------------------
    expiration                Optional integer. This is the length of time in minutes that a token
                              requested through this login process will be valid for.
                              Example: 1440 is one day. The Default is 60.
    ---------------------     --------------------------------------------------------------------
    all_ssl                   Optional boolean. If True, all calls will be made over HTTPS instead
                              of HTTP. The default is False.
    ---------------------     --------------------------------------------------------------------
    portal_connection         Optional string. This is used when a site is federated. It is the
                              ArcGIS Online or Portal GIS object used.
    ---------------------     --------------------------------------------------------------------
    initialize                Optional boolean. If True, the object will attempt to reach out to
                              the URL resource and populate at creation time. The default is False.
    =====================     ====================================================================

    """
    _url = None
    _con = None
    _json_dict = None
    _json = None
    _catalog = None
    _sitemanager = None
    #----------------------------------------------------------------------
    def __init__(self,
                 url,
                 gis=None,
                 **kwargs):
        """Constructor"""
        if gis is None and len(kwargs) > 0:
            if 'baseurl' not in kwargs:
                kwargs['baseurl'] = url
            gis = ServerConnection(**kwargs)
        initialize = kwargs.pop('initialize', False)
        super(Server, self).__init__(gis=gis,
                                     url=url,
                                     initialize=initialize,
                                     **kwargs)

        self._catalog = kwargs.pop('servicesdirectory', None)
        if not url.lower().endswith('/admin'):
            url = "%s/admin" % url
        self._url = url

        #else:
        #    raise ValueError("You must provide either a GIS or login credentials to use this object.")
        if hasattr(gis, '_con'):
            self._con = gis._con
        elif hasattr(gis, '_portal'):
            self._con = gis._portal._con
        elif isinstance(gis, (_ArcGISConnection,
                              ServerConnection)):
            self._con = gis
        else:
            raise ValueError("Invalid gis Type: Must be GIS/ServicesDirectory Object")
        if initialize:
            self._init(self._con)
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def publish_sd(self,
                    sd_file,
                    folder=None):
        """
        Publishes a service definition file to ArcGIS Server.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        sd_file                Required string. The service definition file to be uploaded and published.
        ------------------     --------------------------------------------------------------------
        folder                 Optional string. The folder in which to publish the service definition
                               file to.  If this folder is not present, it will be created.  The
                               default is None in which case the service definition will be published
                               to the System folder.
        ==================     ====================================================================

        :return:
           A boolean indicating success (True) or failure (False).
        """
        import json
        if sd_file.lower().endswith('.sd') == False:
            return False
        catalog = self._catalog
        if 'System' not in catalog.folders:
            return False
        if folder and \
           folder.lower() not in [f.lower() for f in catalog.folders]:
            self.services.create_folder(folder)
        service = catalog.get(name="PublishingTools", folder='System')
        if service is None:
            service = catalog.get(name="PublishingToolsEx", folder='System')
        if service is None:
            return False
        status, res = self._uploads.upload(path=sd_file, description="sd file")
        if status:
            uid = res['item']['itemID']
            if folder:
                config = self._uploads._service_configuration(uid)
                if 'folderName' in config:
                    config['folderName'] = folder
                res = service.publish_service_definition(in_sdp_id=uid,
                                                         in_config_overwrite=json.dumps(config))
            else:
                res = service.publish_service_definition(in_sdp_id=uid)
            return True
        return False
    #----------------------------------------------------------------------
    @staticmethod
    def _create(url,
                username,
                password,
                config_store_connection,
                directories,
                cluster=None,
                logs_settings=None,
                run_async=False,
                **kwargs):
        """
        This is the first operation that you must invoke when you install
        ArcGIS Server for the first time. Creating a new site involves:

          -Allocating a store to save the site configuration
          -Configuring the server machine and registering it with the site
          -Creating a new cluster configuration that includes the server
           machine
          -Configuring server directories
          -Deploying the services that are marked to auto-deploy

        Because of the sheer number of tasks, it usually takes some time
        for this operation to complete. Once a site has been created,
        you can publish GIS services and deploy them to your server
        machines.

        ======================     ====================================================================
        **Argument**               **Description**
        ----------------------     --------------------------------------------------------------------
        connection
        ----------------------     --------------------------------------------------------------------
        url                        Required string. URI string to the site.
        ----------------------     --------------------------------------------------------------------
        username                   Required string. The name of the administrative account to be used by
                                   the site. This can be changed at a later stage.
        ----------------------     --------------------------------------------------------------------
        password                   Required string. The password to the administrative account.
        ----------------------     --------------------------------------------------------------------
        configStoreConnection      Required string. A JSON object representing the connection to the
                                   configuration store. By default, the configuration store will be
                                   maintained in the ArcGIS Server installation directory.
        ----------------------     --------------------------------------------------------------------
        directories                Required string. A JSON object representing a collection of server
                                   directories to create. By default, the server directories will be
                                   created locally.
        ----------------------     --------------------------------------------------------------------
        cluster                    Optional string. An optional cluster configuration. By default, the
                                   site will create a cluster called 'default' with the first available
                                   port numbers starting from 4004.
        ----------------------     --------------------------------------------------------------------
        logsSettings               Optional string. Optional log settings, see http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#/Log_Settings/02r3000001t6000000/ .
        ----------------------     --------------------------------------------------------------------
        runAsync                   Optional boolean. A flag to indicate if the operation needs to be run
                                   asynchronously.
        ======================     ====================================================================


        =====================     ====================================================================
        **Optional Argument**     **Description**
        ---------------------     --------------------------------------------------------------------
        baseurl                   Optional string. The root URL to a site.
                                  Example: https://mysite.com/arcgis
        ---------------------     --------------------------------------------------------------------
        tokenurl                  Optional string. Used when a site is federated or when the token
                                  URL differs from the site's baseurl.  If a site is federated, the
                                  token URL will return as the Portal token and ArcGIS Server users
                                  will not validate correctly.
        ---------------------     --------------------------------------------------------------------
        username                  Optional string. The login username for BUILT-IN security.
        ---------------------     --------------------------------------------------------------------
        password                  Optional string. A secret word or phrase that must be used to gain
                                  access to the account above.
        ---------------------     --------------------------------------------------------------------
        key_file                  Optional string. The path to PKI key file.
        ---------------------     --------------------------------------------------------------------
        cert_file                 Optional string. The path to PKI cert file.
        ---------------------     --------------------------------------------------------------------
        proxy_host                Optional string. The web address to the proxy host.

                                  Example: proxy.mysite.com
        ---------------------     --------------------------------------------------------------------
        proxy_port                Optional integer. The port where the proxy resides on, default is 80.
        ---------------------     --------------------------------------------------------------------
        expiration                Optional integer. This is the length of time a token is valid for.
                                  Example 1440 is one week. The Default is 60.
        ---------------------     --------------------------------------------------------------------
        all_ssl                   Optional boolean. If True, all calls will be made over HTTPS instead
                                  of HTTP. The default is False.
        ---------------------     --------------------------------------------------------------------
        portal_connection         Optional string. This is used when a site is federated. It is the
                                  ArcGIS Online or Portal GIS object used.
        ---------------------     --------------------------------------------------------------------
        initialize                Optional boolean. If True, the object will attempt to reach out to
                                  the URL resource and populate at creation time. The default is False.
        =====================     ====================================================================

        :return:
           Success statement.
        """
        url = url + "/createNewSite"
        params = {
            "f" : "json",
            "cluster" : cluster,
            "directories" : directories,
            "username" : username,
            "password" : password,
            "configStoreConnection" : config_store_connection,
            "logSettings" : logs_settings,
            "runAsync" : run_async
        }
        con = ServerConnection(**kwargs)
        return con.post(path=url,
                        postdata=params)
    #----------------------------------------------------------------------
    def _join(self, admin_url, username, password):
        """
        The Join Site operation is used to connect a server machine to an
        existing site. This is considered a 'push' mechanism, in which a
        server machine pushes its configuration to the site. For the
        operation to be successful, you need to provide an account with
        administrative privileges to the site.
        When an attempt is made to join a site, the site validates the
        administrative credentials, then returns connection information
        about its configuration store back to the server machine. The
        server machine then uses the connection information to work with
        the configuration store.
        If this is the first server machine in your site, use the Create
        Site operation instead.


        ======================     ====================================================================
        **Argument**               **Description**
        ----------------------     --------------------------------------------------------------------
        admin_url                  Required string. The site URL of the currently live site. This is
                                   typically the Administrator Directory URL of one of the server
                                   machines of a site.
        ----------------------     --------------------------------------------------------------------
        username                   Required string. The name of the administrative account for this site.
        ----------------------     --------------------------------------------------------------------
        password                   Required string. The password to the administrative account.
        ======================     ====================================================================


        :return:
           Success statement.
        """
        url = self._url + "/joinSite"
        params = {
            "f" : "json",
            "adminURL" : admin_url,
            "username" : username,
            "password" : password
        }
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def _delete(self):
        """
        Deletes the site configuration and releases all server resources.
        This is an unrecoverable operation. This operation is well suited
        for development or test servers that need to be cleaned up
        regularly. It can also be performed prior to uninstall. Use caution
        with this option because it deletes all services, settings, and
        other configurations.
        This operation performs the following tasks:
          - Stops all server machines participating in the site. This in
            turn stops all GIS services hosted on the server machines.
          - All services and cluster configurations are deleted.
          - All server machines are unregistered from the site.
          - All server machines are unregistered from the site.
          - The configuration store is deleted.

        :return:
           Success statement.
        """
        url = self._url + "/deleteSite"
        params = {
            "f" : "json"
        }
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def _export(self, location=None):
        """
        Exports the site configuration to a location you specify as input
        to this operation.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        location               Optional string. A path to a folder accessible to the server where
                               the exported site configuration will be written. If a location is
                               not specified, the server writes the exported site configuration
                               file to directory owned by the server and returns a virtual path
                               (an HTTP URL) to that location from where it can be downloaded.
        ==================     ====================================================================


        :return:
           Success statement.
        """
        url = self._url + "/exportSite"
        params = {
            "f" : "json"
        }
        if location is not None:
            params['location'] = location
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def _import_site(self, location):
        """
        This operation imports a site configuration into the currently
        running site. Importing a site means replacing all site
        configurations (including GIS services, security configurations,
        and so on) of the currently running site with those contained in
        the site configuration file you supply as input. The input site
        configuration file can be obtained through the exportSite
        operation.
        This operation will restore all information included in the backup,
        as noted in exportSite. When it is complete, this operation returns
        a report as the response. You should review this report and fix any
        problems it lists to ensure your site is fully functioning again.
        The importSite operation lets you restore your site from a backup
        that you created using the exportSite operation.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        location               Required string. A file path to an exported configuration or an ID
                               referencing the stored configuration on the server.
        ==================     ====================================================================


        :return:
           A report.
        """
        url = self._url + "/importSite"
        params = {
            "f" : "json",
            "location" : location
        }
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def _upgrade(self, run_async=False):
        """
        This is the first operation that must be invoked during an ArcGIS
        Server upgrade. Once the new software version has been installed
        and the setup has completed, this operation will be available. A
        successful run of this operation will complete the upgrade of
        ArcGIS Server.

        .. note::
            **caution** If errors are returned with the upgrade operation,
            you must address the errors before you may continue. For example,
            if you encounter an error about an invalid license, you will need
            to re-authorize the software using a valid license and you may
            then retry this  operation.

            This operation is available only when a server machine is
            currently being upgraded. It will not be available after a
            successful upgrade of a server machine.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        run_async              Optional boolean. A flag to indicate if the operation needs to be run
                               asynchronously. The default value is False.
        ==================     ====================================================================


        :return:
           Success statement.
        """
        url = self._url + "/upgrade"
        params = {
            "f" : "json",
            "runAsync" : run_async
        }
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    @property
    def _public_key(self):
        """Gets the public key."""
        url = self._url + "/publicKey"
        params = {
            "f" : "json",
        }
        return self._con.get(path=url,
                             params=params)
    #----------------------------------------------------------------------
    @property
    def machines(self):
        """
        Gets the list of server machines registered with the site.
        This resource represents a collection of all the server machines that
        have been registered with the site. It other words, it represents
        the total computing power of your site. A site will continue to run
        as long as there is one server machine online.
        For a server machine to start hosting GIS services, it must be
        grouped (or clustered). When you create a new site, a cluster called
        'default' is created for you.
        The list of server machines in your site can be dynamic. You can
        register additional server machines when you need to increase the
        computing power of your site or unregister them if you no longer
        need them.
        """
        if self.resources is None:
            self._init()
        if isinstance(self.resources, list) and \
           'machines' in self.resources:
            url = self._url + "/machines"
            return _machines.MachineManager(url,
                                            gis=self._con,
                                            initialize=False)
        else:
            return None
    #----------------------------------------------------------------------
    @property
    def datastores(self):
        """
        Gets the information about the data holdings of the server.
        Data items are used by ArcGIS for Desktop and other clients
        to validate data paths referenced by GIS services.
        You can register new data items with the server by using the
        Register Data Item operation. Use the Find Data Items operation to
        search through the hierarchy of data items.
        A relational data store type represents a database platform that
        has been registered for use on a portal's hosting server by the
        ArcGIS Server administrator. Each relational data store type
        describes the properties ArcGIS Server requires in order to connect
        to an instance of a database for a particular platform. At least
        one registered relational data store type is required before client
        applications such as Insights for ArcGIS can create Relational
        Database Connection portal items.
        The Compute Ref Count operation counts and lists all references to
        a specific data item. This operation helps you determine if a
        particular data item can be safely deleted or refreshed.
        """
        if self.properties is None:
            self._init()
        if isinstance(self.resources, list) and \
           "data" in self.resources:
            url = self._url + "/data"
            return _data.DataStoreManager(url=url,
                                          gis=self._con)
        else:
            return None
    #----------------------------------------------------------------------
    @property
    def _info(self):
        """
        A read-only resource that returns meta information about the server.
        """
        url = self._url + "/info"
        return _info.Info(url=url,
                          gis=self._con,
                          initialize=True)
    #----------------------------------------------------------------------
    @property
    def site(self):
        """
        Gets the site's collection of server resources. This collection
        includes server machines that are installed with ArcGIS Server,
        including GIS services, data and so on. The site resource also
        lists the current version of the software.
        When you install ArcGIS Server on a server machine for the first
        time, you must create a new site. Subsequently, newer server
        machines can join your site and increase its computing power. Once
        a site is no longer required, you can delete the site, which will
        cause all of the resources to be cleaned up.
        """
        if self._sitemanager is None:
            self._sitemanager = SiteManager(self)
        return self._sitemanager
    #----------------------------------------------------------------------
    @property
    def _clusters(self):
        """Gets the clusters functions if supported in resources."""
        if self.properties is None:
            self._init()
        if isinstance(self.resources, list) and \
           "clusters" in self.resources:
            url = self._url + "/clusters"
            return _clusters.Cluster(url=url,
                                     gis=self._con,
                                     initialize=True)
        else:
            return None
    #----------------------------------------------------------------------
    @property
    def services(self):
        """
        Gives the administrator access to the services on ArcGIS Server as a
        ServerManager Object.
        """
        if self.resources is None:
            self._init()
        if isinstance(self.resources, list) and \
           'services' in self.resources:
            url = self._url + "/services"
            return _services.ServiceManager(url=url,
                                            gis=self._con,
                                            initialize=True,
                                            sm=self)
        else:
            return None
    #----------------------------------------------------------------------
    @property
    def usage(self):
        """
        Gets the collection of all the usage reports created
        within your site. The Create Usage Report operation lets you define
        a new usage report.
        """
        if self.resources is None:
            self._init()
        if isinstance(self.resources, list) and \
           'usagereports' in self.resources:
            url = self._url + "/usagereports"
            return _usagereports.ReportManager(url=url,
                                              gis=self._con,
                                              initialize=True)
        else:
            return None
    #----------------------------------------------------------------------
    @property
    def _kml(self):
        """Gets the KML functions for a server."""
        url = self._url + "/kml"
        return _kml.KML(url=url,
                        gis=self._con,
                        initialize=True)
    #----------------------------------------------------------------------
    @property
    def logs(self):
        """
        Gives users access the ArcGIS Server's logs and lets
        administrators query and find errors and/or problems related to
        the server or a service.

        Logs are the records written by the various components of ArcGIS
        Server. You can query the logs and change various log settings.
        **Note**
        ArcGIS Server Only
        """
        if self.resources is None:
            self._init()
        if isinstance(self.resources, list) and \
           'logs' in self.resources:
            url = self._url + "/logs"
            return _logs.LogManager(url=url,
                                    gis=self._con,
                                    initialize=True)
        else:
            return None
    #----------------------------------------------------------------------
    @property
    def _security(self):
        """Gets an object to work with the site security."""
        if self.resources is None:
            self._init()
        if isinstance(self.resources, list) and \
           "security" in self.resources:
            url = self._url + "/security"
            return _security.Security(url=url,
                                      gis=self._con,
                                      initialize=True)
        else:
            return None
    #----------------------------------------------------------------------
    @property
    def users(self):
        """Gets operations to work with users."""
        return self._security.users
    #----------------------------------------------------------------------
    @property
    def content(self):
        """
        Gets the Services Directory which can help you discover information about
        services available on a particular server. A service represents a
        local GIS resource whose functionality has been made available on
        the server to a wider audience. For example, an ArcGIS Server
        administrator can publish an ArcMap document (.mxd) as a map
        service. Developers and clients can display the map service and
        query its contents.

        The Services Directory is available as part of the REST services
        infrastructure available with ArcGIS Server installations. It
        enables you to list the services available, including secured
        services when you provide a proper login. For each service, a set
        of general properties are displayed. For map services, these
        properties include the spatial extent, spatial reference
        (coordinate system) and supported operations. Layers are also
        listed, with links to details about layers, which includes layer
        fields and extent. The Services Directory can execute simple
        queries on layers.

        The Services Directory is also useful for finding information about
        non-map service types. For example, you can use the Services
        Directory to determine the required address format for a geocode
        service, or the necessary model inputs for a geoprocessing service.
        """
        from .. import ServicesDirectory
        if self._catalog is None:
            url = self._url.lower().replace("/admin", "")
            self._catalog = ServicesDirectory(url=url,
                                              con=self._con)
        return self._catalog
    #----------------------------------------------------------------------
    @property
    def system(self):
        """
        Provides access to common system configuration settings.
        """
        if self.resources is None:
            self._init()
        if isinstance(self.resources, list) and \
           "system" in self.resources:
            url = self._url + "/system"
            return _system.SystemManager(url=url,
                                         gis=self._con,
                                         initialize=True)
        else:
            return None
    #----------------------------------------------------------------------
    @property
    def _uploads(self):
        """Gets an object to work with the site uploads."""
        if self.resources is None:
            self._init()
        if isinstance(self.resources, list) and \
           "uploads" in self.resources:
            url = self._url + "/uploads"
            return _uploads.Uploads(url=url,
                                    gis=self._con,
                                    initialize=True)
        else:
            return None
    #----------------------------------------------------------------------
    @property
    def _mode(self):
        """Gets the class that works with Mode."""
        if self.resources is None:
            self._init()
        if isinstance(self.resources, list) and \
           'mode' in self.resources:
            url = self._url + "/mode"
            return _mode.Mode(url=url,
                              gis=self._con,
                              initialize=True)
        return None


########################################################################
class SiteManager(object):
    """
    A site is a collection of server resources. This collection includes
    server machines that are installed with ArcGIS Server, including GIS
    services, data and so on. The site resource also lists the current
    version of the software.
    When you install ArcGIS Server on a server machine for the first time,
    you must create a new site. Subsequently, newer server machines can
    join your site and increase its computing power. Once a site is no
    longer required, you can delete the site, which will cause all of
    the resources to be cleaned up.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    server                 Required string. The arcgis.gis.server object.
    ==================     ====================================================================


    """
    _sm = None
    #----------------------------------------------------------------------
    def __init__(self, server, initialize=False):
        """Constructor"""
        self._sm = server
        isinstance(self._sm, SiteManager)

        if initialize:
            self._sm._init()
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._sm._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._sm._url)
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """Gets the site properties. """
        return self._sm.properties
    #----------------------------------------------------------------------
    @staticmethod
    def create(username,
               password,
               config_store_connection,
               directories,
               cluster=None,
               logs_settings=None,
               run_async=False,
               **kwargs):
        """
        This is the first operation that you must invoke when you install
        ArcGIS Server for the first time. Creating a new site involves:

          -Allocating a store to save the site configuration
          -Configuring the server machine and registering it with the site
          -Creating a new cluster configuration that includes the server
           machine
          -Configuring server directories
          -Deploying the services that are marked to auto-deploy

        Because of the sheer number of tasks, it usually takes some time
        for this operation to complete. Once a site has been created,
        you can publish GIS services and deploy them to your server
        machines.

        ======================     ====================================================================
        **Argument**               **Description**
        ----------------------     --------------------------------------------------------------------
        connection
        ----------------------     --------------------------------------------------------------------
        url                        Required string. URI string to the site.
        ----------------------     --------------------------------------------------------------------
        username                   Required string. The name of the administrative account to be used by
                                   the site. This can be changed at a later stage.
        ----------------------     --------------------------------------------------------------------
        password                   Required string. The password to the administrative account.
        ----------------------     --------------------------------------------------------------------
        configStoreConnection      Required string. A JSON object representing the connection to the
                                   configuration store. By default, the configuration store will be
                                   maintained in the ArcGIS Server installation directory.
        ----------------------     --------------------------------------------------------------------
        directories                Required string. A JSON object representing a collection of server
                                   directories to create. By default, the server directories will be
                                   created locally.
        ----------------------     --------------------------------------------------------------------
        cluster                    Optional string. An optional cluster configuration. By default, the
                                   site will create a cluster called 'default' with the first available
                                   port numbers starting from 4004.
        ----------------------     --------------------------------------------------------------------
        logsSettings               Optional string. Optional log settings, see http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#/Log_Settings/02r3000001t6000000/ .
        ----------------------     --------------------------------------------------------------------
        runAsync                   Optional boolean. A flag to indicate if the operation needs to be run
                                   asynchronously.
        ======================     ====================================================================


        =====================     ====================================================================
        **Optional Argument**     **Description**
        ---------------------     --------------------------------------------------------------------
        baseurl                   Optional string. The root URL to a site.
                                  Example: https://mysite.com/arcgis
        ---------------------     --------------------------------------------------------------------
        tokenurl                  Optional string. Used when a site is federated or when the token
                                  URL differs from the site's baseurl.  If a site is federated, the
                                  token URL will return as the Portal token and ArcGIS Server users
                                  will not validate correctly.
        ---------------------     --------------------------------------------------------------------
        username                  Optional string. The login username for BUILT-IN security.
        ---------------------     --------------------------------------------------------------------
        password                  Optional string. A secret word or phrase that must be used to gain
                                  access to the account above.
        ---------------------     --------------------------------------------------------------------
        key_file                  Optional string. The path to PKI key file.
        ---------------------     --------------------------------------------------------------------
        cert_file                 Optional string. The path to PKI cert file.
        ---------------------     --------------------------------------------------------------------
        proxy_host                Optional string. The web address to the proxy host.

                                  Example: proxy.mysite.com
        ---------------------     --------------------------------------------------------------------
        proxy_port                Optional integer. The port where the proxy resides on, default is 80.
        ---------------------     --------------------------------------------------------------------
        expiration                Optional integer. This is the length of time a token is valid for.
                                  Example 1440 is one week. The Default is 60.
        ---------------------     --------------------------------------------------------------------
        all_ssl                   Optional boolean. If True, all calls will be made over HTTPS instead
                                  of HTTP. The default is False.
        ---------------------     --------------------------------------------------------------------
        portal_connection         Optional string. This is used when a site is federated. It is the
                                  ArcGIS Online or Portal GIS object used.
        ---------------------     --------------------------------------------------------------------
        initialize                Optional boolean. If True, the object will attempt to reach out to
                                  the URL resource and populate at creation time. The default is False.
        =====================     ====================================================================

        :return:
           Success statement.

        """
        return Server._create(username,
                                password,
                                config_store_connection,
                                directories,
                                cluster,
                                logs_settings,
                                run_async)
    #----------------------------------------------------------------------
    def join(self, admin_url, username, password):
        """
        The Join Site operation is used to connect a server machine to an
        existing site. This is considered a 'push' mechanism, in which a
        server machine pushes its configuration to the site. For the
        operation to be successful, you need to provide an account with
        administrative privileges to the site.
        When an attempt is made to join a site, the site validates the
        administrative credentials, then returns connection information
        about its configuration store back to the server machine. The
        server machine then uses the connection information to work with
        the configuration store.
        If this is the first server machine in your site, use the Create
        Site operation instead.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        admin_url              Required string. The site URL of the currently live site. This is
                               typically the Administrator Directory URL of one of the server
                               machines of a site.
        ------------------     --------------------------------------------------------------------
        username               Required string. The name of an administrative account for the site.
        ------------------     --------------------------------------------------------------------
        password               Required string. The password of the administrative account.
        ==================     ====================================================================


        :return:
           A status indicating success or failure.

        """
        return self._sm._join(admin_url, username, password)
    #----------------------------------------------------------------------
    def delete(self):
        """
        Deletes the site configuration and releases all server resources.
        This operation is well suited
        for development or test servers that need to be cleaned up
        regularly. It can also be performed prior to uninstall. Use caution
        with this option because it deletes all services, settings, and
        other configurations.

        This operation performs the following tasks:

          - Stops all server machines participating in the site. This in
            turn stops all GIS services hosted on the server machines.
          - All services and cluster configurations are deleted.
          - All server machines are unregistered from the site.
          - All server machines are unregistered from the site.
          - The configuration store is deleted.

        .. note::
            This is an unrecoverable operation!

        :return:
           A status indicating success or failure.
        """
        return self._sm._delete()
    #----------------------------------------------------------------------
    def export(self, location=None):
        """
        Exports the site configuration to a location you specify as input
        to this operation.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        location               Optional string. A path to a folder accessible to the server
                               where the exported site configuration will be written. If a location
                               is not specified, the server writes the exported site configuration
                               file to directory owned by the server and returns a virtual path
                               (an HTTP URL) to that location from where it can be downloaded.
        ==================     ====================================================================

        :return:
           A status indicating success (along with the folder location) or failure.


        """
        return self._sm._export(location)
    #----------------------------------------------------------------------
    def import_site(self, location):
        """
        This operation imports a site configuration into the currently
        running site. Importing a site means replacing all site
        configurations (including GIS services, security configurations,
        and so on) of the currently running site with those contained in
        the site configuration file you supply as input. The input site
        configuration file can be obtained through the exportSite
        operation.

        This operation will restore all information included in the backup,
        as noted in exportSite. When it is complete, this operation returns
        a report as the response. You should review this report and fix any
        problems it lists to ensure your site is fully functioning again.
        The importSite operation lets you restore your site from a backup
        that you created using the exportSite operation.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        location               Required string. A file path to an exported configuration or an ID
                               referencing the stored configuration on the server.
        ==================     ====================================================================


        :return:
           A status indicating success (along with site details) or failure.
        """
        return self._sm._import_site(location=location)
    #----------------------------------------------------------------------
    def upgrade(self, run_async=False):
        """
        This is the first operation that must be invoked during an ArcGIS
        Server upgrade. Once the new software version has been installed
        and the setup has completed, this operation will be available. A
        successful run of this operation will complete the upgrade of
        ArcGIS Server.


        .. note::
            If errors are returned with the upgrade operation, you must address
            the errors before you can continue. For example, if you encounter
            an error about an invalid license, you will need to re-authorize
            the software using a valid license and you may then retry this
            operation.

            This operation is available only when a server machine is currently
            being upgraded. It will not be available after a successful upgrade
            of a server machine.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        run_async              Required string. A flag to indicate if the operation needs to be run
                               asynchronously. The default value is False.
        ==================     ====================================================================


        :return:
           A status indicating success or failure.

        """
        return self._sm._upgrade(run_async)
    #----------------------------------------------------------------------
    @property
    def public_key(self):
        """Gets the public key."""
        return self._sm._public_key



"""
The root of all OAuth2 resources and operations.
"""
import os
import json
########################################################################
class OAuth(object):
    """
    The root of all OAuth2 resources and operations.
    """
    _gis = None
    _con = None
    _portal = None
    _url = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        self._url = url
        self._gis = gis
        if self._gis is not None:
            self._con = gis._con
        self._portal = gis._portal
    #----------------------------------------------------------------------
    def authorize(self,
                  client_id,
                  response_type,
                  redirect_uris,
                  client_secret=None,
                  state=None,
                  expiration=None,
                  display=None,
                  locale=None,
                  persist=True,
                  ssl=True):
        """
        The Authentication topic describes the overall OAuth2 authentication
        flow. This topic describes the user authorization step of that
        flow.
        Apps that support user logins use OAuth2 to allow users to log in
        to the ArcGIS platform via the app.
        User logins using the OAuth2-based ArcGIS APIs are based on the
        application guiding the user to log in to the platform via a login
        page hosted on the ArcGIS platform. The /oauth2/authorize endpoint
        represents that login page. The login page renders an HTML form for
        the user to enter their credentials.
        The user authentication workflow starts with the authorization
        step. Apps need to direct the browser to this URL. client_id,
        response_type, and redirect_uri are required parameters. There are
        other optional parameters as well, and they;'re described below.
        The response_type parameter determines the type of grant - implicit
        or authorization. A response_type of token implies implicit grant
        and code implies authorization code grant.
        Implicit grants are typically used by JavaScript applications, and
        they complete the flow in a single step. The end result of
        successful authentication is an access_token that's delivered to
        the specified redirect_uri in the URL fragment. See the Response
        section for details.
        Authorization grants are used by mobile, desktop, and server-side
        applications, and they complete the flow in two steps.
        Authorization represents the first step of that flow. Successful
        authorization yields an authorization code that's delivered to the
        specified redirect_uri as a query parameter. See the Response
        section for details. The second step of the flow requires
        exchanging the authorization code obtained in the first step for an
        access_token. This is accomplished by accessing the token endpoint
        with a grant_type of authorization_code.

        Parameters:
         :client_id: The ID of the registered application. Also referred to
          as APPID.
         :response_type: The type of grant - implicit or authorization.
          Values: token, code
          token implies implicit grant and code implies authorization code
          grant.
         :redirect_uris:The URI where the access_token or authorization
          code will be delivered upon successful authorization. The URI
          must match one of the URIs specified during app registration,
          otherwise authorization will be rejected. If registered, a
          special value of urn:ietf:wg:oauth:2.0:oob can also be specified
          for authorization grants. This will result in the authorization
          code being delivered to a portal URL (/oauth2/approval). This
          value is typically used by applications that don't have a web
          server or a custom URI scheme where the code can be delivered.
         :client_secret: The secret of the registered application. Also
          referred to as APPSECRET.
         :state: An opaque value used by applications to maintain state
          between authorization requests and responses. The state, if
          specified, will be delivered back to the redirect_uri as is.
          Applications can use this parameter to correlate the
          authorization request sent with the received response.
         :expiration: The requested validity in minutes of access_token for
          implicit grants or refresh_token for authorization grants.
          For implicit grants, the default validity of access_tokens is 2
          hours. The expiration parameter, if specified, overrides the
          validity period up to a max of 2 weeks (i.e., 20160 minutes).
          For authorization grants, the default validity of access_tokens
          is 30 minutes and refresh_tokens is 2 weeks. The expiration
          parameter, if specified, overrides the validity period of
          refresh_tokens. A permanent refresh_token can be requested by
          specifying expiration=-1.
          Note that org admins can specify the max validity period of
          tokens for their org that supercedes the expiration parameter.
         :display: The template used to render the login page.
          Based on the client platform, applications can choose one of the
          supported templates to render the login page. If not specified,
          the default template is used.
          Values: default, iframe, win8
         :locale: The locale assumed to render the login page.
          Applications can pass the user's locale using this parameter. The
          login page will be rendered using the language corresponding to
          that locale. If not specified, the locale will be determined
          based on the org's setting or on the incoming request.
          Example: locale=fr
         :persist: If true, implies that the user had checked "Keep me
          signed in" when signing into ArcGIS Online.
         :ssl: If true, implies that the user belongs to an ssl-only
          organization
        """
        url = "%s/authorize" % self._url
        params = {
            "f" : "json",
            "redirect_uri" : redirect_uris,
            "response_type" : response_type,
            "client_id" : client_id
        }
        if ssl:
            params['ssl'] = ssl
        if persist:
            params['persist'] = persist
        if locale:
            params['locale'] = locale
        if display:
            params['display'] = display
        if expiration:
            params['expiration'] = expiration
        if state:
            params['state'] = state
        if client_secret:
            params['client_secret'] = client_secret
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def apps(self, client_id):
        """
        An app registered with the portal. An app item can be registered by
        invoking the register app operation. Every registered app gets an
        App ID and App Secret which in OAuth speak are known as client_id
        and client_secret respectively.
        The app owner has access to the registered app resource. This would
        include the organization administrator as well.

        Parameters:
         :client_id: The ID of the registered application. Also referred to
          as APPID.
        """
        url = "%s/apps/%s" % (self._url, client_id)
        params = {"f" : "json"}
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def register_device(self, client_id, expiration=None):
        """
        Registers a device, like mobile phone with a client id to access a
        given Portal or ArcGIS Online Organization.

        Parameters:
         :client_id:The ID of the registered application. Also referred to
          as APPID.
         :expiration: The requested validity in minutes of access_token for
          implicit grants or refresh_token for authorization grants.
          For implicit grants, the default validity of access_tokens is 2
          hours. The expiration parameter, if specified, overrides the
          validity period up to a max of 2 weeks (i.e., 20160 minutes).
          For authorization grants, the default validity of access_tokens
          is 30 minutes and refresh_tokens is 2 weeks. The expiration
          parameter, if specified, overrides the validity period of
          refresh_tokens. A permanent refresh_token can be requested by
          specifying expiration=-1.
          Note that org admins can specify the max validity period of
          tokens for their org that supercedes the expiration parameter.
        """
        url = "%s/registerDevice" % self._url
        params = {
            "f" : "json",
            "client_id" : client_id
        }
        if expiration:
            params['expiration'] = expiration
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def token(self,
              client_id,
              grant_type,
              redirect_uri,
              code=None,
              refresh_token=None,
              client_secret=None):
        """
        TODO: Update Docs
        """
        params = {
            "f" : "json",
            "client_id" : client_id,
            "grant_type" : grant_type
        }

    #----------------------------------------------------------------------
    def register_app(self,
                     item,
                     redirect_uris,
                     app_type="browser"):
        """
        The register app operation (POST only) registers an app item with
        the portal. App registration results in an APPID and APPSECRET
        (also known as client_id and client_secret in OAuth speak,
        respectively) being generated for that app. Upon successful
        registration, a Registered App type keyword gets appended to the
        app item.
        Available to the item owner.

        Parameters:
         :item: arcgis.gis.Item object
         :redirect_uris:The URIs where the access_token or authorization
          code will be delivered upon successful authorization. The
          redirect_uri specified during authorization must match one of the
          registered URIs, otherwise authorization will be rejected.
          A special value of urn:ietf:wg:oauth:2.0:oob can also be
          specified for authorization grants. This will result in the
          authorization code being delivered to a portal URL
          (/oauth2/approval). This value is typically used by apps that
          don't have a web server or a custom URI scheme where the code can
          be delivered.
         :app_type:The type of app that was registered indicating whether
          it's a browser app, native app, server app, or a multiple
          interface app.
          Values: browser | native | server| multiple
        """
        url = "%s/registerApp" % self._url
        itemid = item.id
        params = {
            "f" : "json",
            "itemid" : itemid,
            "appType" : app_type,
            "redirect_uris" : redirect_uris
        }
        return self._con.post(path=url,
                              postdata=params)





"""
These tools are used to identify areas that meet a number of different criteria you specify.

find_similar_locations finds locations most similar to one or more reference locations based on criteria you specify.
"""
import json as _json

import logging as _logging
import arcgis as _arcgis
from arcgis.features import FeatureSet as _FeatureSet
from arcgis.geoprocessing._support import _execute_gp_tool
from ._util import _id_generator, _feature_input, _set_context, _create_output_service

_log = _logging.getLogger(__name__)

# url = "https://dev003153.esri.com/gax/rest/services/System/GeoAnalyticsTools/GPServer"

_use_async = True

def geocode_locations(input_layer,
                      country=None,
                      category=None,
                      include_attributes=True,
                      locator_parameters=None,
                      output_name=None,
                      geocode_service=None,
                      geocode_parameters=None,
                      gis=None):
    """
    The Geocode Locations task geocodes a table from a big data file share. The task uses a geocode
    utility service configured with your portal.

    For more information on setting up a geocoding service see:
    http://server.arcgis.com/en/portal/latest/administer/windows/configure-portal-to-geocode-addresses.htm

    ==========================   ===============================================================
    **Argument**                 **Description**
    --------------------------   ---------------------------------------------------------------
    input_layer                  required Layer, URL, Item of address locations to geocode.
    --------------------------   ---------------------------------------------------------------
    country                      optional string.  If all your data is in one country, this helps
                                 improve performance for locators that accept that variable.
    --------------------------   ---------------------------------------------------------------
    category                     optional string. Enter a category for more precise geocoding
                                 results, if applicable. Some geocoding services do not support
                                 category, and the available options depend on your geocode service.
    --------------------------   ---------------------------------------------------------------
    include_attributes           optional boolean. A boolean value to return the output fields
                                 from the geocoding service in the results.
    --------------------------   ---------------------------------------------------------------
    locator_parameters           optional dictionary. Additional parameters specific to your
                                 locator.
    --------------------------   ---------------------------------------------------------------
    output_name                  optional string, The task will create a feature service of the
                                 results. You define the name of the service.
    --------------------------   ---------------------------------------------------------------
    geocode_service              optional string or Geocoder.  URL endpoint of the Geocoding
                                 Service of GeoCoder object. If none is provided, the service
                                 will use the first geocoder registered with portal that has
                                 batch enabled.
    --------------------------   ---------------------------------------------------------------
    geocode_parameters           optional dictionary.  This includes parameters that help parse
                                 the input data, as well the field lengths and a field mapping.
                                 This value is the output from the AnalyzeGeocodeInput tool
                                 available on your server designated to geocode. It is important
                                 to inspect the field mapping closely and adjust them accordingly
                                 before submitting your job, otherwise your geocoding results may
                                 not be accurate. It is recommended to use the output from
                                 AnalyzeGeocodeInput and modify the field mapping instead of
                                 constructing this JSON by hand.

                                 **Values**

                                 **field_info** - A list of triples with the field names of your input
                                 data, the field type (usually TEXT), and the allowed length
                                 (usually 255).
                                 Example: [['ObjectID', 'TEXT', 255], ['Address', 'TEXT', 255],
                                          ['Region', 'TEXT', 255], ['Postal', 'TEXT', 255]]

                                 **header_row_exists** - Enter true or false.

                                 **column_names** - Submit the column names of your data if your data
                                 does not have a header row.

                                 **field_mapping** - Field mapping between each input field and
                                 candidate fields on the geocoding service.
                                 Example: [['ObjectID', 'OBJECTID'], ['Address', 'Address'],
                                          ['Region', 'Region'], ['Postal', 'Postal']]
    --------------------------   ---------------------------------------------------------------
    gis                          optional GIS, the GIS on which this tool runs. If not
                                 specified, the active GIS is used.
    ==========================   ===============================================================


    :returns: Feature Layer

    """
    from arcgis.features.layer import Layer
    from arcgis.gis import Item
    from arcgis.geocoding._functions import Geocoder
    kwargs = locals()
    tool_name = "GeocodeLocations"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json"
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value
    if output_name is None:
        output_service_name = 'Geocoding_Results_' + _id_generator()
        output_service_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    if isinstance(input_layer, str):
        input_layer = {'url' : input_layer}
    elif isinstance(input_layer, Item):
        input_layer = input_layer.layers[0]._lyr_dict
        if 'type' in input_layer:
            input_layer.pop('type')
    elif isinstance(input_layer, Layer):
        input_layer = input_layer._lyr_dict
        if 'type' in input_layer:
            input_layer.pop('type')
    elif isinstance(input_layer, dict) and \
         not "url" in input_layer:
        raise ValueError("Invalid Input: input_layer dictionary" + \
                         " must have format {'url' : <url>}")
    elif isinstance(input_layer, dict) and "url" in input_layer:
        pass
    else:
        raise ValueError("Invalid input_layer input. Please pass an Item, " + \
                         "Big DataStore Layer or Big DataStore URL to geocode.")

    if geocode_service is None:
        for service in gis.properties.helperServices.geocode:
            if 'batch' in service and service['batch'] == True:
                geocode_service_url = service["url"]
                break
        if geocode_service_url is None:
            raise ValueError("A geocoder with batch enabled must be configured" + \
                             " with this portal to use this service.")
        params['geocode_service_url'] = geocode_service_url
    elif isinstance(geocode_service, Geocoder):
        geocode_service = geocode_service.url
        params['geocode_service_url'] = geocode_service_url
    elif isinstance(geocode_service, str):
        params['geocode_service_url'] = geocode_service
    else:
        raise ValueError("geocode_service_url must be a string or GeoCoder")

    if geocode_parameters is None:
        from arcgis.geoprocessing._tool import Toolbox
        analyze_geocode_url = gis.properties.helperServices.asyncGeocode.url
        tbx = Toolbox(url=analyze_geocode_url, gis=gis)
        geocode_parameters = tbx.analyze_geocode_input(input_table=input_layer,
                                                       geocode_service_url=geocode_service_url)
        params['geocode_parameters'] = geocode_parameters
    output_service = _create_output_service(gis, output_name,
                                            output_service_name, 'Geocoded Locations')
    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "geocode_service_url": (str, "geocodeServiceURL"),
        "geocode_parameters": (str, "geocodeParameters"),
        "country": (str, "sourceCountry"),
        "category": (str, "category"),
        "include_attributes" : (bool, "includeAttributes"),
        "locator_parameters" : (str, "locatorParameters"),
        "output_name": (str, "outputName"),
        "output": (_FeatureSet, "output"),
        "context": (str, "context")
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        res = _execute_gp_tool(gis, tool_name, params, param_db,
                               return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise
    return

def detect_incidents(input_layer,
                     track_fields,
                     start_condition_expression,
                     end_condition_expression=None,
                     output_mode="AllFeatures",
                     output_name=None,
                     gis=None):
    """
    The Detect Incidents task works with a time-enabled layer of points,
    lines, areas, or tables that represents an instant in time. Using
    sequentially ordered features, called tracks, this tool determines
    which features are incidents of interest. Incidents are determined by
    conditions that you specify. First, the tool determines which features
    belong to a track using one or more fields. Using the time at each
    feature, the tracks are ordered sequentially and the incident condition
    is applied. Features that meet the starting incident condition are
    marked as an incident. You can optionally apply an ending incident
    condition; when the end condition is true, the feature is no longer
    an incident. The results will be returned with the original features
    with new columns representing the incident name and indicate which
    feature meets the incident condition. You can return all original
    features, only the features that are incidents, or all of the features
    within tracks where at least one incident occurred.

    For example, suppose you have GPS measurements of hurricanes every 10
    minutes. Each GPS measurement records the hurricane's name, location,
    time of recording, and wind speed. Using these fields, you could create
    an incident where any measurement with a wind speed greater than 208
    km/h is an incident titled Catastrophic. By not setting an end
    condition, the incident would end if the feature no longer meets the
    start condition (wind speed slows down to less than 208).

    Using another example, suppose you were monitoring concentrations of a
    chemical in your local water supply using a field called
    contanimateLevel. You know that the recommended levels are less than
    0.01 mg/L, and dangerous levels are above 0.03 mg/L. To detect
    incidents, where a value above 0.03mg/L is an incident, and remains an
    incident until contamination levels are back to normal, you create an
    incident using a start condition of contanimateLevel > 0.03 and an end
    condition of contanimateLevel < 0.01. This will mark any sequence where
    values exceed 0.03mg/L until they return to a value less than 0.01.

    ==========================   ===============================================================
    **Argument**                 **Description**
    --------------------------   ---------------------------------------------------------------
    input_layer                  required FeatureSet, The table, point, line or polygon features
                                 containing potential incidents.
    --------------------------   ---------------------------------------------------------------
    track_fields                 required string, The fields used to identify distinct tracks.
                                 There can be multiple track_fields.
    --------------------------   ---------------------------------------------------------------
    start_condition_expression   The condition used to identify incidents. If there
                                 is no endConditionExpression specified, any feature
                                 that meets this condition is an incident. If there
                                 is an end condition, any feature that meets the
                                 start_condition_expression and does not meet the
                                 end_condition_expression is an incident.
                                 The expressions are Arcade expressions.
    --------------------------   ---------------------------------------------------------------
    end_condition_expression     The condition used to identify incidents. If there is
                                 no endConditionExpression specified, any feature that
                                 meets this condition is an incident. If there is an
                                 end condition, any feature that meets the
                                 start_condition_expression and does not meet the
                                 end_condition_expression is an incident. This is an
                                 Arcade expression.
    --------------------------   ---------------------------------------------------------------
    output_mode                  optional string, default value is AllFeatures.  Determines
                                 which features are returned. Two modes are available:

                                 - AllFeatures - All of the input features are returned.
                                 - Incidents - Only features that were found to be incidents
                                   are returned.
    --------------------------   ---------------------------------------------------------------
    output_name                  optional string, The task will create a feature service of the
                                 results. You define the name of the service.
    --------------------------   ---------------------------------------------------------------
    gis                          optional GIS, the GIS on which this tool runs. If not
                                 specified, the active GIS is used.
    ==========================   ===============================================================

    :returns:
       Output feature layer item
    """
    kwargs = locals()
    tool_name = "DetectIncidents"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json"
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Detect_Incidents_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Detect Track Incidents')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "track_fields": (str, "trackFields"),
        "start_condition_expression": (str, "startConditionExpression"),
        "end_condition_expression": (str, "endConditionExpression"),
        "output_mode": (str, "outputMode"),
        "output_name": (str, "outputName"),
        "output": (_FeatureSet, "output"),
        "context": (str, "context")
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

    return

def find_similar_locations(
    input_layer,
    search_layer,
    analysis_fields,
    most_or_least_similar = """MostSimilar""",
    match_method = """AttributeValues""",
    number_of_results = 10,
    append_fields = None,
    output_name = None,
    gis = None):
    """

    Based on criteria you specify, find similar locations by measuring the similarity of locations in your candidate search layer to one or more reference locations.

    For example

    * Find the ten most similar stores by examining the number of employees and the annual sales.

    * Find the 100 most similar cities by examining the relationship between population, annual growth, and tax revenue.


   Parameters:

   input_layer: Input Layer (feature layer). Required parameter.

   search_layer: Search Layer (feature layer). Required parameter.

   analysis_fields: Analysis Fields (str). Required parameter.

   most_or_least_similar: Most Or Least Similar (str). Required parameter.
      Choice list:['MostSimilar', 'LeastSimilar', 'Both']

   match_method: Match Method (str). Required parameter.
      Choice list:['AttributeValues', 'AttributeProfiles']

   number_of_results: Number Of Results (int). Required parameter.

   append_fields: Fields To Append To Output (str). Optional parameter.

   output_name: Output Features Name (str). Required parameter.

   gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   output - Output feature layer Item


    """
    kwargs = locals()

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Similar Locations_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Find Similar Locations')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "search_layer": (_FeatureSet, "searchLayer"),
        "analysis_fields": (str, "analysisFields"),
        "most_or_least_similar": (str, "mostOrLeastSimilar"),
        "match_method": (str, "matchMethod"),
        "number_of_results": (int, "numberOfResults"),
        "append_fields": (str, "appendFields"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        _execute_gp_tool(gis, "FindSimilarLocations", params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

find_similar_locations.__annotations__ = {
    'most_or_least_similar': str,
    'match_method': str,
    'number_of_results': int,
    'append_fields': str,
    'output_name': str}




"""
The arcgis.geocoding module provides types and functions for geocoding, batch geocoding and reverse geocoding.

Geocoders can find point locations of addresses, business names, and so on.
The output points can be visualized on a map, inserted as stops for a route,
or loaded as input for spatial analysis. It is also used to generate
batch results for a set of addresses, as well as for reverse geocoding,
i.e. determining the address at a particular x/y location.
"""

from ._functions import *
